#ifndef ARRAY
#define ARRAY

#include <stdio.h>
#include <stdarg.h>

/*
 *  A simple multi-dimensional array class. 
 *
 *  ... could probably be improved.  Probably slow as.  Suspect a real
 *  whizz could make it aesthetically snappier too - there's a fair
 *  bit of cut and paste code repetition which I can't quite see how
 *  to hide...
 *
 *  WARNING:   It is very easy to incorrectly initialise these suckers.
 *             The first Parameter MUST be the total rank of the array.
 *             Hence a simple 4x3 array is declared as: 
 *                CBArray Small(2,4,3)
 */


class CBArray
{
private:

   int  Rank;
   long size;

   float* base;
   int*  IndexRank; 

public:

   CBArray()
   {
       Rank = 0;
       size = 0;
       IndexRank = NULL;
       base      = NULL; 
   }

   CBArray(int rank, ...)
   {
      va_list ap;
      va_start(ap, rank);
   
      Rank = rank; 
      IndexRank = new int[rank];
      size = 1;

      for (int i=0; i<rank; i++)
      {
          IndexRank[i] = va_arg(ap, int); 
          size *= IndexRank[i];
      } 
 
      base = new float[size];

      for (long i=0; i<size; i++)
          base[i]=0;

      va_end(ap);
   }

   CBArray(int rank, int* index) 
   {
      Rank = rank; 

      IndexRank = new int[rank];
      size = 1;

      for (int i=0; i<rank; i++)
      {
          IndexRank[i] = index[i]; 
          size *= IndexRank[i];
      } 
 
      base = new float[size];

      for (long i=0; i<size; i++)
          base[i]=0;
   }

   ~CBArray()
   {
       delete IndexRank;
       delete base;
   }

   void CBArray::Resize(int rank, int * index)
   {
      Rank = rank;

      delete IndexRank;
      IndexRank = new int[rank];
      size = 1;
      
      for (int i=0; i<rank; i++)
      {
          IndexRank[i] = index[i];
          size *= IndexRank[i];
      }

      delete base;
      base = new float[size];

      for (long i=0; i<size; i++)
          base[i]=0;
   }

   void CBArray::Resize(int rank, ...)
   {
      va_list ap;
      va_start(ap, rank);
   
      Rank = rank; 
      IndexRank = new int[rank];
      size = 1;

      for (int i=0; i<rank; i++)
      {
          IndexRank[i] = va_arg(ap, int); 
          size *= IndexRank[i];
      } 
 
      base = new float[size];

      for (long i=0; i<size; i++)
          base[i]=0;

      va_end(ap);
   }

   void CBArray::PrintError(int indexval, int* index)
   {
       cerr << "\nBad CBArray index "<<indexval<<": is "<<index[indexval];
       cerr << " max was: "<<IndexRank[indexval]<<"\n";
       cerr << "(call was array[ ";
       for (int j=0; j<Rank; j++) cerr << index[j] <<" ";
       cerr << "]\n";
   }

   float& CBArray::operator()(int* index)
   {
      int  temp;
      long pos    = 0;
      long offset = 1;

      for (int i=0; i<Rank; i++)
      {
          temp = index[i];
          if (temp>=IndexRank[i])
          {
              PrintError(i, index);

              return base[0]; 
          }
          pos += offset*temp; 
          offset *= IndexRank[i]; 
      }

      if (pos>size) {cerr << "inconceivable!\n"; pos=0;}

      return base[pos];
   }

   float& CBArray::operator()(int A, ... )
   {
      va_list ap;
      va_start(ap, A);

      long pos    = 0;
      int  temp   = A;
      long offset = 1;

      for (int i=0; i<Rank; i++)
      {
          if (temp>=IndexRank[i])
          {
              cerr << "Error in CBArray size " << Rank<<" : " << temp << " larger than or equal to " << IndexRank[i] << " in pos " << i << "\n"; 
              return base[0]; 
          }
          pos += offset*temp; 
          offset *= IndexRank[i]; 
          if (i<Rank-1) temp = va_arg(ap, int); 
      }

      va_end(ap);

      if (pos>size) {cerr << "inconceivable!\n"; pos=0;}

      return base[pos]; 
   }

   friend ostream& operator << (ostream &o, CBArray &A);
   friend ostream& operator << (ostream &o, CBArray *A);
};

ostream& operator << (ostream &o, CBArray &A);
ostream& operator << (ostream &o, CBArray *A);

#endif
