/*
 *	CB2NANO  is a "chris - format" file to nano-simulator struct format 
 *      file converter
 *	"chris - format" is a high level object description language
 *	thingy, designed to make life easy.  It consists of position
 *	declarations of spheres, cylinders and helices, and also allows
 *	the "#include"ing of other "chris-format" files.
 *	- in addition, it allows raw text to be entered, which is 
 * 	simply passed through to the output file.
 */

#ifndef CB2NANO
#define CB2NANO

#include <string.h>
#include <iostream.h>
#include <iomanip.h>
#include <fstream.h>
#include <stdlib.h>
#include <stdio.h>

#ifndef NOSGI
#include "math.h"
#else
#include <math.h>
#endif


#include "Filter.h"
#include "pt.h"

#define START " "
#define TWOPI 6.28318


class cb2nano : public Filter
{
    public:

    ofstream* Nano_file;
    pt OffsetAll;
    pt RotateAll;
    static pt PtArray[];
    static int NoPts;
    static int level;  // count include file recursions;

    int ObjectCount, TotalObjects;


    char *OutputFileName;

    cb2nano(); 

    void CleanPtArray();
    void Make_sphere(char *Name, pt Center);
    virtual void FilterFtn(ifstream* InFile, ofstream* OutFile);
    void Make_cylinder(char *Name, float Radius, pt End1, pt End2);
    void Make_truncus(char *Name, float Radius1, float Radius2, pt End1, pt End2);
    void Make_cone(float Radius, pt End1, pt End2);
    void Make_helix(char *Name, float Period, int Strands, float Unit_length, float Radius, float S_radius, pt End1, pt End2);
    void Make_hollow_cylinder(char *Name, float Radius, float Radius2, pt End1, pt End2);
    virtual void HelpMsg();

    //Following redundancy is 'CC' silliness.
    int  Read_file(char *file_name);
    int  Read_file(char *file_name, pt C);
    int  Read_file(char *file_name, pt C, pt D);

    void Make_block(pt End1, pt End2, pt Center, pt Direction);
    void Make_block(pt End1, pt End2, pt Center, pt Direction, pt Dir2);
    void Make_pipe(char *Name, float Radius1, float Radius2, pt End1, pt End2);
    void Make_flex(float Radius1, float Radius2, pt End1, pt End2, float Grad1, float Grad2);
    void Make_sheath(char *Name, float Radius1, float Radius2, pt End1, pt End2, float Grad1, float Grad2, float Thickness, float S_angle, float F_angle);
    void Make_Actin(char *Name, char *text);
    void Make_Actin_blob(char *Name, pt S, pt D, float Size);
    virtual void ParseCommand(char* S);
    void SetOffset(char *S);
    void SetRotate(char *S);
 
    void Write();
    void Write(char *text);
   
};

#endif 


			

