/*
 *	CB2POV  is a "chris - format" file to povray format file converter
 *	"chris - format" is a tres' cunning object description language
 *	thingy, designed to make life easy.  It consists of position
 *	declarations of spheres, cylinders and helices, and also allows
 *	the "#include"ing of other "chris-format" files.
 *	- in addition, all normal povray commands are valid, and are
 * 	simply passed through to the output file.
 */

#ifndef CB2POV
#define CB2POV

#include <string.h>
#include <iostream.h>
#include <iomanip.h>
#include <fstream.h>
#include <math.h>
#include <stdlib.h>

#include "Filter.h"
#include "pt.h"

#define HEADER "#include \"colors.inc\"\n#include \"shapes.inc\"\n#include \"textures.inc\""
#define CAMERA "\ncamera {\n  location <0, 0, 10>\n  look_at <0,0,0>\n}"
#define LIGHTS "light_source { <0, 100, 1000> color rgb <1,1,1>}"
#define TWOPI 6.28318
#define TEXTURE "default { texture { finish {phong 0.9 phong_size 60 metallic}}}"
#define START " "

class cb2pov : public Filter
{
    ofstream* Pov_file;
    pt OffsetAll;
    pt RotateAll;
    static pt PtArray[];
    static int NoPts;
    static int level;  // count include file recursions;
    static char Pigment[];
    static char Finish[];
    static char Normal[];

    void CleanPtArray();
    void End_object();
    void Make_sphere(float Size, pt Center);
    virtual void FilterFtn(ifstream* InFile, ofstream* OutFile);
    void Make_cylinder(float Radius, pt End1, pt End2);
    void Make_truncus(float Radius1, float Radius2, pt End1, pt End2);
    void Make_cone(float Radius, pt End1, pt End2);
    void Make_helix(float Period, int Strands, float Unit_length, float Radius, float S_radius, pt End1, pt End2);
    void Make_hollow_cylinder(float Radius, pt End1, pt End2, float Thickness, float S_angle, float F_angle);
    virtual void HelpMsg();

    //Following redundancy is 'CC' silliness.
    int  Read_file(char *file_name);
    int  Read_file(char *file_name, pt C);
    int  Read_file(char *file_name, pt C, pt D);
    void Make_block(pt End1, pt End2, pt Center, pt Direction);
    void Make_block(pt End1, pt End2, pt Center, pt Direction, pt Dir2);
    void Make_pipe(float Radius1, float Radius2, pt End1, pt End2);
    void Make_flex(float Radius1, float Radius2, pt End1, pt End2, float Grad1, float Grad2);
    void Make_game_board(pt End1, pt End2, float Divisions);
    void Make_trapezoid(pt P1, pt P2, pt P3, pt P4);
    void Make_sheath(float Radius1, float Radius2, pt End1, pt End2, float Grad1, float Grad2, float Thickness, float S_angle, float F_angle);
    void Make_Actin(char *text);
    void Make_Actin_blob(pt S, pt D, float Size);
    virtual void ParseCommand(char* S);
    void SetOffset(char *S);
    void SetRotate(char *S);
 
    void Write();
    void Write(char *text);
    void WriteTriangle(pt a, pt b, pt c);
    void WriteCurrentTexture();
    void Write_colour();
   
    void Start_blob(float Threshold);
    void Add_blob(float Size, pt C);
    void End_blob(); 
};

#endif 


			

