#ifndef DIMER
#define DIMER 

/*
 *   TO DO:
 *
 *   The concept of links must be generalised, as must the handling
 *   of binding behaviour.
 *
 */

//#define VERBOSE    // used for verbose debugging statements in log...

#include <math.h>
#include <values.h>
#include <iostream.h>

#ifdef NOSGI
#include <limits.h>
#endif


#include "pt.h"
#include "matrix.h"
#include "link.h"
#include "dataclasses.h"
#include "tubparser.h"
#include "structure.h"
#include "physobject.h"

class Microtubule;

//#define NUMB_DIMERS 500   //Number of dimers in simulation
#define DEF_NUMB_DIMERS 500 //the default number of dimers in the simulation

#define CALCS_PER_FRAME 200 // display each 'n' calculation cycles...

#define REPHOS_CHANCE  10    // n% chance/cyle of GDP -> GTP

#ifdef NOSGI    // Un-be-f**king-leavable!
#define R1  ((float)random()/(float)LONG_MAX)        // random number 0->1
#else
#define R1  ((float)random()/(float)2147483647)        // random number 0->1
#endif

#define Rpm ((R1*2)-1)                              // random # -1 -> 1 
#define R2pm ((R1*4)-2)                             // random # -2 -> 2 
#define RpmS ((R1*Field_Size)-Field_Size/2)         // random gridpos
#define R10  R1*10                                  // random # 0 -> 10 
#define R100 R1*100                                 // random # 0 -> 100
#define Rpm20 (R1*40)-20                            // random # -20 -> 20 
#define RAND_U_VECT ~pt(Rpm,Rpm,Rpm)                // random unit vector
#define RAND_DU_VECT ~pt(R2pm,R2pm,R2pm)            // random double unit vect.

#define TRIPHOS 0
#define DIPHOS 1

// Numbers reduced by .01 to avoid rounding errors
//#define MIN_DIST    1.99    // Min dist 'tween central pts of 2 tubulin proteins
//#define MIN_DIST_SQ 3.99    // Min sqr dist of approach between centers 
//#define MAX_DIST_C_SQ 46.99 // Max sqr dist 'tween centers that allows collision
//#define MAX_I_DIST_SQ 46.99 // The max interaction dist square 'tween dimer centers
//#define MAX_I_DIST    6.99  // The max interaction dist 'twixt dimers

//#define TOP    0
//#define BOTTOM 1
//#define LEFT   2
//#define RIGHT  3
//#define SIDE   4

//#define NOLINKS 4

//#define DONOTHING 1

#define BOUND    -1
#define NOTBOUND  0

class Dimer : public PhysicalObject 
{
public:

// -----------------DATA-----------------------------------------------------


    int bound;

    pt bound_velocity;          // velocity when bound in a mt.

//    CBMatrix Basis;             // The basis vectors, expressed as a matrix 

    ModelData* Model;           // The data for this type of V.protein.

    int NoSites;                // identical (local copy of) to Model->NoSites
    int NoSpheres;              // identical (local copy of) to Model->NoSpheres

//    pt center;                  // the co-ordinate center of the object

    float BoundRad;             // radius smallest enclosing sphere. Copy from DataModel. 

//    float rms_dp;               // rms delta x/y/z in nanometers. Copy from DataModel.

    pt LocalMomentOfInertia;    // says it all really.

    int State;                  // Current State of object. 

    int FalsePos;               // Dimer temp. moved "past" field boundaries 

    int TagNo;                  // give the dimers names for debugging

    Link LinkNo[MAXSITES];       // List of the current links of this dimer - one per site

    pt    SpherePos[MAXSPHERES]; // actual positions of sphere centers.
    float SphereRad[MAXSPHERES]; // redundant copy of sphere radii

    Microtubule *Owner;          // the microtubule this dimer is part of (if any).

    int RandomEventFlag;         // True if a random event could change the current state

    int tempFlag;                // Temporary Flag - used occasionally in multi-dimer procs.

// ----------------METHODS----------------------------------------------------

    Dimer();
    Dimer(ModelData* M); 
    ~Dimer();

    void CheckLinkTypes(char* Msg);

    void  SetData(ModelData *M);
    void  SetData(ModelData *M, PosData *P);
    void  unbind();
    void  add(pt A);

    void  jump();  // randomly move a dimer
    void  spin();                // randomly orient dimer
    void  bump();  // randomly move *and* randomly spin a dimer. 

    void  bump_bound_dimer();
    void  rotate_bound_dimer(pt Centroid, pt AngVel);
    void  rotate_bound_dimer(pt Centroid, pt axis, float angle);
    void  clear();
    void  free();
    void  printlinks(int PrintAllLinks=TRUE);

    void  SetTwistMatrices();
    void  SetSpheres();
    void  SetLinks();

    void  SetOwner(Microtubule* M);
    int   GetBound(int ignore_errors = FALSE);
    int   GetFullyBound();
    int   CheckLinks();

    void  SetToMyCoords(Dimer *D);
    void  SetToCoords(CBMatrix B, pt Center);
//    int   LinkUp(Dimer *D, Link *L1, Link *LD);
//    void  break_link(Dimer *D);
    void  RecalculateDetails();
    void  AddOffset(pt offset);
    float BreakageChance();

    void  FreeBySelf();

    void  FalsifyPosition(pt offset);
    void  DeFalsifyPosition();
//    void  DeFalsifyPosition(pt offset);
    void  Normalise();
    void  SetState(int NewState);
    void  CheckStateOnBreak(Link* from);
    void  CheckStateOnBind(Link* from, Link* to);
    void  CheckStateOnRandomEvent();

    void  UnLinkAll();

    Link* FindLink(Dimer *D);

// -------------STATIC DATA-----------------------------------------------------

    static int NoObjects;
    static int Field_Size;
    static int Half_Field_Size;
    static long NoCollisions;  // used in collision_test mode only.
// ------------STATIC METHODS----------------------------------------------------

    static void Report();
    static int CheckFieldSize(int FieldSize);
    static TubParser *Init(int NoDimers, int FieldSize, char *PDDFStr = NULL, char *StructureString = NULL);
    static void InitRandomBindStrengths();
    static void ClearAll();
    static void CheckAll(char* Msg);

    static pt    Sproing(pt A, pt B, float spring_const);
    static void  CalculateBindingMatrix();

    static int   probftn(float Strength, float Dist, Dimer* A, Dimer* B);
    static float min_contact_dist(Dimer *a, Dimer *b); //, pt *diff = NULL, int GetDiff = FALSE);
    static float min_contact_dist_sq(Dimer *a, Dimer *b);//, pt *diff = NULL, int GetDiff = FALSE);

//    static void  BreakLink(Dimer *A, Dimer *B);
    static void  Dump_cells();
    static void  Agitate();
    static void  Check_binding();
    static int   CheckInteraction(Dimer *A, Dimer *B);
    static int   Interact(Dimer *A, Dimer *B);
    static void  GetLinksAndDistSq(Dimer *A, Dimer *B, Link **Alink, Link **Blink, float &Dist);
    static int   LinksAlreadyBound(Link *Alink, Link *Blink);
    static float GetBindingStrength(Link *Alink, Link *Blink);
    static void  Verify_spacing();
    static pt    RepulsionVector(Dimer *A, Dimer *B);
    static void  Repulse(Dimer *A, Dimer *B);
    static void  Random_events();
    static void  Integrity_check();
    static int   Collision(Dimer *A, Dimer *B);


    static int   LinkUp(Dimer *A, Dimer *B, Link *LA, Link *LB);
    static int   UnLink(Dimer *A, Dimer *B);
    static int   Aggregate(Dimer *A, Dimer *B);

//DEBUG ftns.

    static int    Get_pos(Dimer *P);
    static void   Dump_dimers();

    static int   CoupleLinks(Dimer *A, Dimer *B, Link *LA, Link *LB);
    static int   Align(Dimer *A, Dimer *B, Link *LA, Link *LB);

    static Dimer*  Dimers;

    friend ostream& operator << (ostream &o, Dimer &D); 
    friend ostream& operator << (ostream &o, Dimer *D); 

    // If you ask me, C++ static constants aren't worth the hasstle.

};
    ostream& operator << (ostream &o, Dimer &D); 
    ostream& operator << (ostream &o, Dimer *D); 

#endif
