#include "gridcell.h" 
#include "assert.h"

GridCell GridCell::Grid[GRID_SIZE][GRID_SIZE][GRID_SIZE];
int GridCell::Field_Size;
int GridCell::Half_Field_Size;

int DEBUGHACK = FALSE;

/*
 *  The Grid Cell class 
 *
 *  - handles the 3D array of buckets the dimers are tossed
 * into.  It has methods to deal with the whole set of dimers,
 * moving, handling collisions, and generally doing all the
 * large scale functionality.
 *
 * We use buckets so we don't have to test all possible dimer
 * interactions.
 *
 * Hmmm... possibly should OOP-ify this a little more...
 */

void GridCell::Init(int fieldsize)
{
   Field_Size = fieldsize;
   Half_Field_Size = Field_Size / 2;
   int CellSize = Field_Size / (int) GRID_SIZE;
   GridCell *G;
   int NoCells = 0;

   for (int i=0; i<GRID_SIZE; i++)
     for (int j=0; j<GRID_SIZE; j++)
       for (int k=0; k<GRID_SIZE; k++)
       {
           G = &Grid[i][j][k];
           G->i_pos = i;
           G->j_pos = j;
           G->k_pos = k;
           if (i==0||i==GRID_SIZE-1) G->IsBorder = TRUE;
           if (j==0||j==GRID_SIZE-1) G->IsBorder = TRUE;
           if (k==0||k==GRID_SIZE-1) G->IsBorder = TRUE;
           G->FalsePos = FALSE;
           G->x_low  = i*CellSize - Half_Field_Size; 
           G->x_high = (i+1)*CellSize - Half_Field_Size;
           G->y_low  = j*CellSize - Half_Field_Size;
           G->y_high = (j+1)*CellSize - Half_Field_Size;
           G->z_low  = k*CellSize - Half_Field_Size;
           G->z_high = (k+1)*CellSize - Half_Field_Size;
           G->TagNo = NoCells++;
       }
}


/*
 *   This runs a ftn over all the possible interactions between a 
 *   list of dimers (except that a given dimer does not interact 
 *   with itself).  It returns TRUE if any call to the parameter
 *   ftn returns TRUE, otherwise FALSE;
 */

int GridCell::ProcessDimerList( int (*ftn)(Dimer *A, Dimer *B) )
{
    register i,j;
    int Ret_value = FALSE;

    for (i=0; i<No; i++)
        for (j=i+1; j<No; j++)
        {
            if ((*ftn)(List[i], List[j]) == TRUE)
               Ret_value = TRUE; 
        }
    return Ret_value;
} 


/*
 *   This runs a ftn over all possible interactions between
 *   two lists of dimers.  A given list does not interact
 *   with itself.  It returns "TRUE" if any call to the 
 *   parameter ftn returns TRUE, otherwise FALSE; 
 *
 *   This proc handles dimers on either side of the "wrap-around"
 *   boundary.
 *
 *   - This proc is static.
 */

int GridCell::ProcessDimerLists( GridCell *a, GridCell *b, int (*ftn)(Dimer *A, Dimer *B) )
{
    register i,j;
    int Ret_value = FALSE; 
    pt offsetA = pt(0);
    pt offsetB = pt(0);

    // Handle the boundary problem by "wrapping" cells around, if necessary

    if ((a->i_pos==0 && b->i_pos==GRID_SIZE_M1)||
        (a->j_pos==0 && b->j_pos==GRID_SIZE_M1)||
        (a->k_pos==0 && b->k_pos==GRID_SIZE_M1))
    {
        offsetB = pt(((b->i_pos==GRID_SIZE_M1)?-Field_Size:0), 
                    ((b->j_pos==GRID_SIZE_M1)?-Field_Size:0), 
                    ((b->k_pos==GRID_SIZE_M1)?-Field_Size:0));
        b->FalsifyCellPosition(offsetB); 
    }

    if ((b->i_pos==0 && a->i_pos==GRID_SIZE_M1)||
        (b->j_pos==0 && a->j_pos==GRID_SIZE_M1)||
        (b->k_pos==0 && a->k_pos==GRID_SIZE_M1))
    {
        offsetA = pt(((a->i_pos==GRID_SIZE_M1)?-Field_Size:0), 
                    ((a->j_pos==GRID_SIZE_M1)?-Field_Size:0), 
                    ((a->k_pos==GRID_SIZE_M1)?-Field_Size:0));
        a->FalsifyCellPosition(offsetA); 
    }

    for (i=0; i<a->No; i++)
        for (j=0; j<b->No; j++)
        {
            if ((*ftn)(a->List[i], b->List[j])==TRUE)
                Ret_value = TRUE;
        }

if ((a->FalsePos==TRUE)&&(b->FalsePos==TRUE)) cout << "\n*Double Whammy*\n";

    if (a->FalsePos==TRUE)              // if cells *were* wrapped around, put
        a->DeFalsifyCellPosition(offsetA);   // them back to their original positions.
    if (b->FalsePos==TRUE) 
        b->DeFalsifyCellPosition(offsetB);

    return Ret_value;
}


/*
 *   Handles possible binding between two dimers
 */

int GridCell::Interaction(Dimer* A, Dimer* B)
{
//VEM EVIL
    if (A->rms_dp==0 && B->rms_dp==0) return FAIL;

    if (A->Owner!=NULL&B->Owner!=NULL)
        if (A->Owner == B->Owner) 
        {
            return FAIL;   // don't do anything, everything is just dandy.
        }

    if (pt::fast_dist(A->center, B->center)<MAXINTERACTIONRANGE)
    {  
        if (Dimer::Interact(A, B)==BOUND)
        {
            return OK;
        }
        
    }
    return FAIL;
}


/*
 * Runs any dimer/dimer interactions within a single grid cell
 */

void GridCell::SelfInteraction()
{
    if (No>1) ProcessDimerList(Interaction);
}


/*
 *   Runs any dimer/dimer interactions between dimers in adjoining grid cells 
 */

void GridCell::Interaction(GridCell* cellA, GridCell* cellB)
{
    if (cellA->No>0&&cellB->No>0)
    {
        ProcessDimerLists(cellA, cellB, Interaction);
    }
}


/*
 *   Checks for collisions between dimers in a single grid cell.  
 *   => Returns TRUE if a collision.
 */

int GridCell::SelfCollisions()
{
    if (No<=1) return FALSE; 

    int collision = FALSE;
    int CollisionOccured = FALSE;
    int overload = 0;

    do
    {
        collision = ProcessDimerList(Dimer::Collision);
        if (collision==TRUE) CollisionOccured = TRUE;

        if (overload++>10) 
        {
            cout << "Too Many Reps in SelfCollisions()\n"; 
            return FALSE;
        }
    } while (collision == TRUE);
 
    if (CollisionOccured)
    {
        CleanUpStrayDimers();  // handle dimers that may have bounced out of the cell 
        Mark();
    }
    return (CollisionOccured);

}

int GridCell::CleanUpStrayDimers()
{
    int overload = 0;  // No of Cyles through this procedure.

    while (CheckIfDimersLeftCell())  // see if dimers are still within this cell
    {
        Dimer *D = GetStrayDimer();        // get a dimer that moves past cell bounds
        DeleteDimer(D);                    // remove dimer from this cell
        LoadDimerIntoCell(D);              // ... and put it where it now belongs.
        GetCellForPos(D->center)->Mark();  // mark the new cell as changed

//DEBUGHACK = TRUE;
//GridCell *G = GetCellForPos(D->center);
//cout << "Dimer "<<D->TagNo<<" has left the cell: placed in : " << G->TagNo<<"  ("<<D->center<<")\n";
//DEBUGHACK = FALSE;

        if (overload++>10)  // this might need to be changed for crowded runs.
        {

            cout << "Doofus up in CleanUpStrayDimers(): Bailing "<<overload<<" ";
            cout << "( " << D->center << " )\n";
            return FALSE;
        }
    }
    return TRUE;

}

Dimer* GridCell::GetStrayDimer()
{
    for (int i=0; i<No; i++)
        if (DimerWithinCell(List[i])==FALSE)
            return List[i];
    return FALSE;    
}

int GridCell::CheckIfDimersLeftCell()
{
    for (int i=0; i<No; i++)
    {
        if (DimerWithinCell(List[i])==FALSE) 
        {
            return TRUE;
        }
    }
    return FALSE; 
}

int GridCell::DimerWithinCell(Dimer *D)
{
    if (D->center.x>=x_low&&D->center.x<=x_high&&D->center.y>=y_low&&
        D->center.y<=y_high&&D->center.z>=z_low&&D->center.z<=z_high)
    {
        return TRUE;
    }
    else
    {
        return FALSE;
    }
}
/*
 *  Mark all cells adjacent to [i][j][k] as possibly changed, and therefore
 *  in need of re-examination.
 */

void GridCell::Mark()
{
    int I,J,K,ii,jj,kk;

    for (ii=i_pos-1;ii<=i_pos+1;ii++)
        for (jj=j_pos-1;jj<=j_pos+1;jj++)
            for (kk=k_pos-1;kk<=k_pos+1;kk++)
            {
                I=ii; J=jj; K=kk;
                if (I==GRID_SIZE) I=0;             //adjusted for wrap-around 
                if (J==GRID_SIZE) J=0;
                if (K==GRID_SIZE) K=0;
                 if ((I>=0)&&(J>=0)&&(K>=0))
                     Grid[I][J][K].Marked=TRUE;
             }
}


/*
 *  Checks/handles collisions between dimers in adjoining grid cells.
 *  => Returns TRUE if a collision.
 */

int GridCell::Collision(GridCell *A, GridCell *B)
{
    if (B->No==0) return FALSE;

    int collision = FALSE;
    int CollisionOccured = FALSE;
    int overload = 0;
 
    do
    {   
        collision = ProcessDimerLists(A, B, Dimer::Collision);

        if (collision == TRUE) CollisionOccured = TRUE;

        overload++;
        if (overload>10) {cout << "Too Many Reps in Collision()\n"; return FALSE;}
    } while (collision == TRUE);
   
    if (CollisionOccured)
    { 
        A->CleanUpStrayDimers();  // Check if any dimers have left the cell
        B->CleanUpStrayDimers();  // ditto
        A->Mark();
        B->Mark();
    }

    return CollisionOccured;
}


/*
 *   Cycle through all possible cell-cell interactions.
 *
 *   Evil Code.  Should be prettified.
 */

void GridCell::GlobalInteraction()
{
    int i,j,k,ii,jj,kk,I,J,K;
    GridCell *A, *B;

    for (i=0; i<GRID_SIZE; i++) for (j=0; j<GRID_SIZE; j++) for (k=0; k<GRID_SIZE; k++)
    {
        A = &GridCell::Grid[i][j][k];

        if (A->No>0)
        {
            A->SelfInteraction();  

            //there is no need to check all 26 neighbours...
            for (ii=i;ii<=i+1;ii++) for (jj=j;jj<=j+1;jj++) for (kk=k;kk<=k+1;kk++)
//            for (ii=i;ii<=i+1;ii++) for (jj=j-1;jj<=j+1;jj++) for (kk=k-1;kk<=k+1;kk++)
            {
                I =ii; J=jj; K=kk;

                if (I==GRID_SIZE) I=0;        // the grid cells wrap around: 
                if (J==GRID_SIZE) J=0;
                if (K==GRID_SIZE) K=0;

//                if (((I>=0)&&(J>=0)&&(K>=0)) && ((I!=i)||(J!=j)||(K!=k)))
                if ((I!=i)||(J!=j)||(K!=k))
                {
                    B = &GridCell::Grid[I][J][K];

                    if (B->No>0) 
                    {
                        //Check for interactions between adjacent cells
                        GridCell::Interaction(A,B);
                    }         
                }
            }
       }
    }
}

            
/*
 *    Go through all grid cells.
 *    Find and remove any collisions.
 *    If there *were* collisions, mark surrounding cells as suspect
 *    Go back and check *them* for collisions
 *    Keep going until everything clear
 * 
 *    Evil Code: Should be prettified.  Maybe combined with previous ftn?
 */

void GridCell::CheckSpacing()
{
    int No_Passes = 0;

    while (SingleSpacingPass(!No_Passes)==TRUE)
    {
        No_Passes++;    

        if (No_Passes > 10) 
        {
            cout <<"Hrm. Check Spacing Runaway prob "<<No_Passes<<" times\n"; 
            return;
        }

        if (No_Passes > 3)
            cout << "Hmmm... deepish iteration in Check Spacing..."<<No_Passes<<"\n";
 
        CheckAllBounds();  // (this is expensive, so we don't do it every
        //LoadCells();       // time)
    }
}


/*
 *  Go through all grid cells checking their constituent dimers for collisions with
 *  other dimers.
 */

int GridCell::SingleSpacingPass(int CheckAll)
{
    int i,j,k,ii,jj,kk,I,J,K;
    int Changed = FALSE;
    GridCell *A, *B;

    for (i=0; i<GRID_SIZE; i++) for (j=0; j<GRID_SIZE; j++) for (k=0; k<GRID_SIZE; k++)
    {
        A = &Grid[i][j][k];

//SEEMS DUMB 7/12/97        if (CheckAll||((A->Marked)&&(A->No>0)))

        if ((A->No)>0&&(CheckAll||A->Marked))
        {
            
            if (A->SelfCollisions()==TRUE)     // Check cell for self collisions
            {                            
                Changed =TRUE; 
            }

            //Check cell against "forward" neighbours
//TOO MANY 7/12/97            for (ii=i;ii<=i+1;ii++) for (jj=j-1;jj<=j+1;jj++) for (kk=k-1;kk<=k+1;kk++)
            for (ii=i;ii<=i+1;ii++) for (jj=j;jj<=j+1;jj++) for (kk=k;kk<=k+1;kk++)
            {
                I =ii; J=jj; K=kk;

                if (ii==0&&jj==0&&kk==0) continue;

                if (I==GRID_SIZE) I=0;        // the grid cells wrap around: 
                if (J==GRID_SIZE) J=0;
                if (K==GRID_SIZE) K=0;

                // the following makes sure we're accessing valid grid cells
                if (((I>=0)&&(J>=0)&&(K>=0)) && ((I!=i)||(J!=j)||(K!=k)))
                    if ((J>=0)&&(K>=0)&&(I<GRID_SIZE)&&(J<GRID_SIZE)&&(K<GRID_SIZE))
                        if ((I!=i)||(J!=j)||(K!=k))
                        {
                            B = &Grid[I][J][K];
                            if (B->No==0) continue;

                            if (Collision(A,B)==TRUE) // Check adjoining cells 
                            {
                                Changed=TRUE;    
                            }
                        }
             }
	} 
    }    
    return Changed;
}

void GridCell::LoadCells()
{
    /*
     *  the result of -ve(int)/(int) is implementation dependant:
     *  i.e. -1/5 may be either 0 or -1.  In order to smoothly
     *  allocate our dimers into their grid cells, I would like 
     *  -1/5 to *always* = -1 (regardless of platform) - so I
     *  introduce a NEGative DIVide MODifier to force the result 
     *  to always round to the lowest number...
     */

    //int NEGDIVMOD = -1-(-1/3);

    // Initialize Grid cell counters to zero.
    

    int i,j,k;

    for (i=0; i<GRID_SIZE; i++) for (j=0; j<GRID_SIZE; j++) for (k=0; k<GRID_SIZE; k++)
    {
         Grid[i][j][k].No     = 0;
         Grid[i][j][k].Marked = FALSE;
    }

    /*
     *    Place dimers in their grid cells.
     *    (nb. - dimers that wander off are wrapped around.) 
     */


    for (i=0; i<Dimer::NoObjects; i++)
    {
        LoadDimerIntoCell(&Dimer::Dimers[i]);
    }
}


GridCell* GridCell::GetCellForPos(pt P)
{
    const int CellSize = Field_Size/(int)GRID_SIZE;

// May need to go-faster all this.  Possibly some cunning binary
// arithmatic? restrict CellSize/FieldSize to int powers of two?

    int a = (int)(P.x/CellSize ) + (HALF_GRID_SIZE) + (P.x<0?NEGDIVMOD:0);
    int b = (int)(P.y/CellSize ) + (HALF_GRID_SIZE) + (P.y<0?NEGDIVMOD:0);
    int c = (int)(P.z/CellSize ) + (HALF_GRID_SIZE) + (P.z<0?NEGDIVMOD:0);

/*
 *  Error check that dimer hasn't wondered off.  If it has,
 *  return NULL and let the caller cope with it.
 */

    if ((a>=GRID_SIZE)||(a<0) || (b>=GRID_SIZE)||(b<0) || (c>=GRID_SIZE)||(c<0))
    {
//cout << "Stray dimer ["<<a<<"]["<<b<<"]["<<c<<"] == pos: " << P <<"\n";

        if (P.x == Field_Size) a = GRID_SIZE;
        if (P.y == Field_Size) b = GRID_SIZE;
        if (P.z == Field_Size) c = GRID_SIZE;
      

        if ((a>=GRID_SIZE)||(a<0) || (b>=GRID_SIZE)||(b<0) || (c>=GRID_SIZE)||(c<0))
        {
            return NULL;
        }
        else
            return &Grid[a][b][c];

    }
/*if (DEBUGHACK)
{
    cout << " CELL: ["<<a<<"] ["<<b<<"] ["<<c<<"] no: ";
    cout <<Grid[a][b][c].TagNo <<"\n";
}
*/
    return &Grid[a][b][c];
}

void GridCell::LoadDimerIntoCell(Dimer *D)
{
    GridCell* G = GetCellForPos(D->center);

    if (G==NULL)  // Shouldn't happen often at all. 
                  // Usually dimers that have only just slipped over the edge
    {
        D->Normalise(); // adjust position.
        G = GetCellForPos(D->center);
        if (G==NULL) // Still?  Very Bad things. Bail Out Hard.
                     // usually D->center is <NaN,NaN,NaN>.
        {
            cout << TheProgram.NoCyclesCompleted << " Error -> dimer in bad cell.  clearing dimer "<<D->TagNo;
            if (D->Owner!=NULL) cerr << " owned by " << D->Owner->TagNo;
            cout << "\n";
            cout << " Pos: ("<<D->TagNo<<") " << D->center << "\n";
            delete(D->Owner);  // Bad MT.  Die.
            D->clear();
            G = GetCellForPos(D->center);

        }
    }

    if (G->No<GRID_CELL_MAX)
         G->Add(D);
    else
    {
         cerr<< "dimer["<<D->TagNo<<"] overflow! "<<G->i_pos<<" "<<G->j_pos<<" "<<G->k_pos<<" : "<<D->center.x<<" "<<D->center.y<<" "<<D->center.z<<"   (field size = " << Field_Size << ") ["<<GRID_CELL_MAX<<"] \n";
    }

    if (G->DimerWithinCell(D)==FALSE)
    {
cout << "Dimer Lost : " << D->TagNo << "  " << D->center << "     NEGDIVMOD " << NEGDIVMOD <<"\n";
cout << "Doofus Up in Load Cells [" << G->i_pos <<"] ["<<G->j_pos<<"] ["<<G->k_pos<<"]  ... \n";
cout << " Pos: " <<G->x_low<<":"<<G->x_high<<" "<<G->y_low<<":"<<G->y_high;
cout << " "<<G->z_low<<":"<<G->z_high<<"\n\n"; 


    } 
}


/*
 *   Remove a dimer from a grid cell.
 */

void GridCell::DeleteDimer(Dimer *D)
{
    int Replacing = FALSE;
    for (int i=0; i<No; i++)
    {
        if (D==List[i]) 
        {
            Replacing = TRUE;
//            cout << "Found Dimer " << D->TagNo << "; removing from " <<TagNo<<" \n";
        }

        if (Replacing == TRUE)
        {
            if (i<(No-1))
                List[i]=List[i+1];
        }
    }

    if (Replacing==FALSE)
    {
        cerr << " Error: Cannot find Dimer " << D->TagNo << " in Grid Cell ";
        cerr << "["<<i_pos<<"] ["<<j_pos<<"] ["<<k_pos<<"]\n";
    }
    else
    {
        No--;
    }
}

/*
void GridCell::DumpGridNumbs()
{
    if (TheProgram.silent) return;

    ERRORLOG << "DumpGridNumbs\n\nGrid Size " << GRID_SIZE<<"\n";
    int i,j,k;
    for (i=0; i<GRID_SIZE; i++)
    {
        for (j=0; j<GRID_SIZE; j++)
        {
            for (k=0; k<GRID_SIZE; k++)
            {
                if (Grid[i][j][k].No == 0)
                    ERRORLOG << ". ";
                else
                    ERRORLOG << Grid[i][j][k].No << " ";
            }
            ERRORLOG << "\n";
        }
        ERRORLOG << i<<" ...\n";
    }
    ERRORLOG << "End Dump\n";
}

void GridCell::DumpCells()
{
    if (TheProgram.silent) return;

    //ERRORLOG << "DumpCells : Grid A\n";
    int i,j,k,l;
    for (i=0; i<GRID_SIZE; i++)
        for (j=0; j<GRID_SIZE; j++)
            for (k=0; k<GRID_SIZE; k++)
                for (l=0; l<Grid[i][j][k].No; l++)
                    ERRORLOG << "\n i j k  # "<<i<<" "<<j<<" "<<k<<" "<<GridCell::Grid[i][j][k].List[l];

    ERRORLOG << "\nEnd Cell Dump\n";
}
*/

/*
 *  Checks that all the dimers are within the simulation field 
 *  -> and forces them in if they ain't (by wrapping the simulation
 *     field).
 */

void GridCell::CheckAllBounds()
{
    for (int i=0; i<Dimer::NoObjects; i++)
    {
//VEM_EVIL
        if (Dimer::Dimers[i].rms_dp != 0) 
            CheckBounds(&Dimer::Dimers[i]);
    }
}


/*
 *  Checks that a given dimer is within the simulation field,
 *  and moves them there if they aren't
 */

void GridCell::CheckBounds(Dimer *D)
{
    D->Normalise();
}



void GridCell::DebugCheckAllDimers()
{
    for (int i=0; i<Dimer::NoObjects; i++)
    {
        if (TheProgram.OutOfBounds(Dimer::Dimers[i].center))
        {
//            if (pt::abs(Dimer::Dimers[i].center)>200)

            {
            cout << TheProgram.NoCyclesCompleted << "Dimer " << Dimer::Dimers[i].TagNo << " bad\n " << Dimer::Dimers[i]<<"\n";
            }
        }
    }
}

/*
 *  Moves all the elements in a cell to a new position so they
 *  can be meaningfully checked for valid interactions across the
 *  (wrap-around) simulation field boundary. 
 */

void GridCell::FalsifyCellPosition(pt offset)
{
    if (No!=0&&offset!=0)  
    {
        FalsePos = TRUE;              // flag cell for later "defalsification"
        for (int i=0; i<No; i++)
        {
            Dimer* D = List[i];
            D->FalsifyPosition(offset);
        }
    }
    else if (offset==0)
        cout << "Unnecessary Call to FCP by " << i_pos<<","<<j_pos<<","<<k_pos <<" \n ";
}


/*
 *  This returns points to their correct pos.  For efficiency, 
 *  (since we have the variable on hand) we can call this with
 *  the correct offset rather than calling DeFalsifyPosition()
 *  and then Normalise(), which would require a bit more computation.  
 */

void GridCell::DeFalsifyCellPosition(pt offset)
{
    if (No==0||offset==0||FalsePos==FALSE)  
        {cout << "Unnecessary call to DFCP by "<<i_pos<<","<<j_pos<<","<<k_pos<<" " << No << " "<<offset<<" "<<FalsePos<<"\n"; FalsePos = FALSE; return;}
       
    for (int i=0; i<No; i++)
    {
//       cout << " DFCP:"<<List[i]->TagNo<<" ";
//       List[i]->DeFalsifyPosition(offset);
       List[i]->DeFalsifyPosition();
       FalsePos = FALSE;
    } 
}


