#ifndef GRID
#define GRID

#include <iostream.h>
#include <stdlib.h>
#include <stdio.h>

#include "microtubule.h"
#include "dimer.h"
#include "hashdefines.h"


#define CYCLE_ijk_0_TO_SIZE_OF_GRID_M1 for (i=0; i<GRID_SIZE; i++) for (j=0; j<GRID_SIZE; j++) for (k=0; k<GRID_SIZE; k++) 


class GridCell {
public:
    Dimer** List;
    int     No;
    int     TagNo;     // debugging unique i.d.
    int     Marked;
    int     IsBorder;
    int     FalsePos;
    int     i_pos,j_pos,k_pos; 
    int     x_low,x_high,y_low,y_high,z_low,z_high;


    GridCell()        {No=0; List = new Dimer *[GRID_CELL_MAX];
                       Marked = FALSE; IsBorder=FALSE; FalsePos = FALSE;}
    GridCell(int max) {No=0; List = new Dimer *[max];
                       Marked = FALSE; IsBorder=FALSE; FalsePos = FALSE;}

    int Add(Dimer* NewDimer) {if (No<GRID_CELL_MAX) List[No++]=NewDimer; 
                              return (No>GRID_CELL_MAX);}

//----------Functions-------------------------------------------

    int   ProcessDimerList( int (*ftn)(Dimer *A, Dimer *B) );
    void  DeFalsifyCellPosition(pt offset=0);
    void  FalsifyCellPosition(pt offset);
    int   SelfCollisions();
    void  SelfInteraction();
    void  Mark();
    void  DeleteDimer(Dimer *D);
    int   DimerWithinCell(Dimer *D);
    int   CleanUpStrayDimers();
    int   CheckIfDimersLeftCell();
    Dimer* GetStrayDimer();

//----------Static Data-------------------------------------------

    static int Field_Size;
    static int Half_Field_Size;

//----------Static Functions-------------------------------------------

    static void  Init(int fieldsize);
    static int   ProcessDimerLists( GridCell *A, GridCell *B, int (*ftn)(Dimer *A, Dimer *B));
    static void  GlobalInteraction();
    static int   Interaction(Dimer* A, Dimer* B);
    static void  Interaction(GridCell *cellA, GridCell *cellB);
    static int   Collision(GridCell *cellA, GridCell *cellB);
    static void  DumpGridNumbs();
    static void  DumpCells();
    static void  LoadCells();
    static void  LoadDimerIntoCell(Dimer *D);
    static void  CheckSpacing();
    static int   SingleSpacingPass(int CheckAll);
    static void  CheckAllBounds();
    static void  CheckBounds(Dimer *D);
    static void  DebugCheckAllDimers();

    static GridCell* GetCellForPos(pt P);
    static GridCell Grid[GRID_SIZE][GRID_SIZE][GRID_SIZE];
};

#endif 
