#ifndef HASHDEFINES
#define HASHDEFINES

/*********************************************************
 *   Data Model Hash Defines
 *********************************************************
 *
 *   These define the maximum number of attributes the models of the
 *   system can have.  These should be changed if a particularly 
 *   large simulation is to be done, and the program recompiled.
 */

#define MAXMODELS       10  
#define MAXSPHERES      10
#define MAXSTATES       20
#define MAXEVENTS       100
#define MAXSITES        100
#define MAXLINKS        64        // HARD LIMIT: 64 MAY NOT BE EXCEEDED 
#define MAXCOLOURS      50

#define MAXINTERACTIONRANGE 100   // max collision range = 100 nanometers
#define MAXDISTS        22        // granularity of distance measures
#define MAXRADS         41        // granularity of radius measures.

/*********************************************************
 *   Protein constants
 *********************************************************
 *
 *
 */

#define PROTEIN_DENSITY 1   // (in 1e-23*grams/nm**3) - e.g. as water.

/************************************************************************
 * Event hash defines 
 ***********************************************************************/


#define BREAKEVENT  1
#define BINDEVENT   2
#define RANDOMEVENT 3


/************************************************************************
 * Linkage/Site hash defines 
 ***********************************************************************/

#define UNKNOWN -1
#define NONE    -2
#define ALL     -3


/************************************************************************
 * Program parameter defaults
 ***********************************************************************/

#define DEF_MOLARITY    20  // default molarity (used if field size changes)
#define DEF_NUMB_DIMERS 500 //the default number of dimers in the simulation
#define CALCS_PER_FRAME 200 // display each 'n' calculation cycles...
#define DEF_FIELD_SIZE 256


/*************************************************************************
 * Hash defines for GridCells
 ************************************************************************
 *
 * some thinking required.  Don't mess with lightly...
 * This is for region FIELD_SIZE**3, with dimers 4*2*2 units in size.
 */

#define GRID_SIZE 8         //No of cells in a dimension of location grid
#define GRID_SIZE_M1 (GRID_SIZE-1)
#define HALF_GRID_SIZE 4 
#define NEGDIVMOD (-1-(-1/3))
#define GRID_CELL_MAX 512     // max dimers per cell: should work out a physical
                             // packing limit or make dynamic...

/*************************************************************************
 * Physical Constants 
 ************************************************************************/

#define PI 3.1415927
#define BOLTZMANN_CONST 1.38066e-23

/*************************************************************************
 * Program Return Values 
 ************************************************************************/

#define ERROR 1
#define EXIT  1
#define OK    0
#define DONOTHING -2 
#define FAIL  2
#define FALSE 0

/*************************************************************************
 * Simulation Constants 
 ************************************************************************/

#define USEFASTMATHS         // use fast, approximate maths routines


#endif
