#ifndef LINK
#define LINK

#include "matrix.h"
// #include "array.h"
#include "pt.h"
#include "dataclasses.h"
#include "assert.h"
/*
 *  This is basically a sorta wrapper thing for a LinkData type.
 *
 *  The general forms of the various links are read in by the parser,
 *  and loaded into a set of generic links in the TubParser class. 
 *
 *  These Links know what Type of link they are, as well as some local
 *  info (where exactly they are, what they are bound to).
 */

class Dimer;

class Link
{
public:

    LinkData* Type;      // General data for this *type* of link


                         // specific data for this individual link : 

    Dimer    *to;        // the dimer bound to (if any) 
    pt        true_pos;  // and what position.


    float     breakprob;     // current chance of breakage

    // constuctors.

    Link()
    {
        to = NULL; 
        Type = NULL;
        true_pos = pt(0,0,0); 
    } 

    Link(Link* L)
    {
        Type      = L->Type;
        to        = L->to;
        true_pos  = L->true_pos;
        breakprob = L->breakprob;
    }

    Link(LinkData* L) 
    {
        Type = L;
        to = NULL; 
        true_pos = pt(0,0,0); 
    }

    void UpdateType(LinkData* L)
    {
        assert( L!= NULL);
        Type = L; 
    }
  
    void SetBreakProb(Link *L)
    {
        assert (L!=NULL);
        breakprob = Type->BreakProb[L->Type->LinkNo]; 
    }
 
// Hmmm... I wonder if all this overloading will prove confusing...
// (ans: yes.  Kill it.) 
//    Link operator=  (Dimer *To)   {to = To; return *this;}
    Link operator=  (pt P)        {true_pos = P; return *this;}
    Link operator+= (pt P)        {true_pos = true_pos + P; return *this;}
//    int  operator== (Dimer *test) {return (test==to);} 

    //void free()                   {to = NULL;}
    void move(pt P)               {true_pos = Type->pos + P;}    

    void SetTruePos(pt P)         {true_pos = P;}

    int  Bound()                  {return (to!=NULL);}
//    int  CanBind(Link *L)         {return ((Type->BindsTo & L->Type->TypeId)!=FALSE);} 
    int  CanBind(Link *L)         {return Type->CanBind(L->Type);} 
 
    friend ostream& operator << (ostream &o, Link &L);
    friend ostream& operator << (ostream &o, Link *L);
};

ostream& operator << (ostream &o, Link &L);
ostream& operator << (ostream &o, Link *L);

#endif
