#include "tubular.h"
#include "pt.h"
#include "matrix.h"

/*
 *  Given a direction vector (an axis passing through (0,0,0) 
 *  and an angle, this returns a matrix which
 *  will rotate any point around that axis and angle.
 *
 *  e.g. if Axis = pt(1,0,0) (a unit x-vector)
 *       and radians = 3.14159/2 (90 degrees)
 *
 *   Then pt y_vector(0,1,0) * RotationMatrix(Axis, radians) 
 *     would equal minus_z_vector(0,0,-1)
 *
 *   -> Using the left hand, if the thumb is the "z" axis, the forefinger
 *   the "y" axis, and the middle finger x, then the direction of rotation
 *   is a "left hand curl" -> i.e. the direction of the fingers if the
 *   left hand is wrapped around the x-axis, with the thumb now pointing
 *   in the +x direction.
 *
 *   Did that make any sense?
 * 
 */

CBMatrix CBMatrix::RotationMatrix (pt Axis, double radians)
{
/*
 *  The static members mean that if the function is called twice with the
 *  same parameters, it won't bother to repeat the calculation.
 */

    static CBMatrix Rotator;
    static pt P = pt(1,0,0);
    static double theta = 0;

    if ((P!=Axis)||(theta!=radians))
    {

        if (pt::sqr(Axis)!=1) Axis = ~Axis;

        P       = Axis;
        theta   = radians;

        double C = cos(theta);
        double A = 1-C;
        double B = sin(theta);

/*
 *  This could be go-fastered if neccessary...
 */

        Rotator = CBMatrix ((A*P.x*P.x) + C,     A*P.x*P.y + B*P.z, A*P.x*P.z - B*P.y,
                            A*P.y*P.x - B*P.z, (A*P.y*P.y) + C,     A*P.y*P.z + B*P.x,
                            A*P.z*P.x + B*P.y, A*P.z*P.y - B*P.x, (A*P.z*P.z) + C);
    }
//    else
//        cout << "reusing calcs\n";
    return Rotator;
}


/*
 *    Matrix related stuff
 */

CBMatrix CBMatrix::Identity(1,0,0,0,1,0,0,0,1);

ostream & operator << (ostream &o, CBMatrix *M)
{
    o << (*M);
    return o;
}

ostream & operator << (ostream &o, CBMatrix &M)
{
    o << " Matrix (det = "<< M.det() << ")\n";
    if (M.a<0.0001&&M.a>-0.0001) M.a = 0;
    if (M.b<0.0001&&M.b>-0.0001) M.b = 0;
    if (M.c<0.0001&&M.c>-0.0001) M.c = 0;
    if (M.d<0.0001&&M.d>-0.0001) M.d = 0;
    if (M.e<0.0001&&M.e>-0.0001) M.e = 0;
    if (M.f<0.0001&&M.f>-0.0001) M.f = 0;
    if (M.g<0.0001&&M.g>-0.0001) M.g = 0;
    if (M.h<0.0001&&M.h>-0.0001) M.h = 0;
    if (M.i<0.0001&&M.i>-0.0001) M.i = 0;




    o << " [ "<<M.a<<" "<<M.b<<" "<<M.c<<" ]\n";
    o << " [ "<<M.d<<" "<<M.e<<" "<<M.f<<" ]\n";
    o << " [ "<<M.g<<" "<<M.h<<" "<<M.i<<" ]\n";
    if (M.is_identity==TRUE) o << "is identity\n"; else o << "not identity\n";
   
    return o;
}

