#ifndef MICROT
#define MICROT

#include "tubular.h"
#include "dimer.h"
#include "gridcell.h"
#include "physobject.h"

class Microtubule : public PhysicalObject
{
public:

//---------------Data---------------------------------------

    int TagNo;

    Microtubule* Next;
    Microtubule* Back;
//    int CurrentArraySize;
    int ArraySizeMax;
    int NoObjects;
    Dimer **DimerPtArray;
    pt velocity;              // actual change in position in given cycle
//    float rms_dp;             // rms delta position/cycle  (standard dev. of movement)
//    float Mass;               // Mass in arbitrary units for mechanics calcs.

    pt   Centroid;
//    pt   LocalMomentOfInertia;
//    CBMatrix Basis;           // co-ords of local ref. frame.
                              // (global_pt = Basis * local_pt) (?)

    CBMatrix InvBasis;        // co-ords of global frame (in local co-ords)
                              // (local_pt  = InvBasis * global_pt) (?)
//    float Diffusion;

//-----------------Methods----------------------------------

    Microtubule(Dimer *A);
    ~Microtubule();

    int Add(Dimer *D, Dimer *local_dimer);
    int IntegrityCheck();
    int IntegrityCheck(char *Msg);
    int  Move();
    void MoveLinearly(pt velocity);
//  void MoveRotationally(pt AngVel);    //rotate * ang. velocity (local co-ords)
    void Rotate(pt Axis, float Angle); //rotate in absolute co-ords
    int  CheckDimerPastBoundary();
    void DeleteDimer(Dimer *D);
    int  CheckBeforeAdding(Dimer *A, Dimer *B, Link *LinkA, Link *LinkB);
    int  CheckMT();

    void FullPrint(int PrintAllLinks=TRUE);
    void PrintDimerList();

    void StripDimerFromOwner(Dimer *D);
    void RecursivelyAddDimer(Dimer *A, Dimer *B, Link *Alink, Link *Blink);
    void RecursivelyFreeDimer(Dimer *B, Microtubule* OwnerB);

    void InitRotationalData(Dimer *D);
    void AddMass(Dimer *D);
    void DeleteMass(Dimer *D);
    void RecalculateMassDetails();
    float CalculateMass();
    pt    CalculateCentroid(float Mass);
    pt    CalculateMOI(pt Centroid);
    float CalculateRMS_DP(pt MOI);

    pt   MakeLocal(pt p);
    int  CheckDimer(Dimer *D);   
    int  CheckDimerPos(Dimer *D);
    int  CheckLinks();
    int  CheckLinkPair(Link A, Link B);

    int  CheckCrossLinks(Dimer *D);
    void BreakCrossLinks(Dimer *D);
    int  CheckLinkBetween(Dimer *A, Dimer *B);

    int  AddDimerToArray(Dimer *D);
    int  RemoveDimerFromArray(Dimer *D);

    void MoveDimersToCommonRegion();
    void MoveDimersToTruePositions();

    void RecalculateDimerPositions(); 

 
// -----------------Static Data----------------------------------

    static Microtubule* First;
    static Microtubule* Last;

    static int NoMicroTs;

    static long totalmass;
    static long count;
    static long turbidity;
    static float avweight;


// -----------------Class Functions------------------------------

    static void Report();
    static void FinalReport();
    static void CreateNewMicrotubule(Dimer *A, Dimer *B);
    static void MoveAll();
    static void CheckAll(char* Msg);

    //ftns to do with melding two microtubules together.

    static int  Meld(Dimer *A, Dimer *B);
    static void MeldTop(Dimer *A, Dimer *B, pt A_dirtn);
    static void MeldSide(Dimer *A, Dimer *B );
    static void PhysicallyJoin(Dimer *A, Dimer *B, Link *Alink, Link *Blink);
    static int  ProbMeld(Dimer *A, Dimer *B);


    friend ostream& operator << (ostream &o, Microtubule &T);
    friend ostream& operator << (ostream &o, Microtubule *T);


private:

    int CheckArrayInAdd();
};
    ostream& operator << (ostream &o, Microtubule &T);
    ostream& operator << (ostream &o, Microtubule *T);


#endif
