#ifndef PARSER
#define PARSER
#ifndef SHORTSTRING
#define SHORTSTRING
#endif

/*
 *   A parser class - reads a file in, and extracts yummy useful info...
 */

#include "tubular.h"

#include <string.h>
#include <iostream.h>
#include <fstream.h>
#include <stdlib.h>

#define FALSE 0
#define TRUE  1

class ShortString  // Quickly reinvent the wheel.  (I want *my* wheel, dammit!)
{
public:

    char* text; 
  
    static char InputBuffer[255];
 
    ShortString() {text = NULL;}
    ShortString(char *Text) {text = strdup(Text);}
    ShortString(const ShortString *S) {text = strdup(S->text);}
    ShortString(const ShortString &S) {text = strdup(S.text);}
    
    ShortString operator= (const char *Text)     {text = strdup(Text); return this;}
    ShortString operator= (const ShortString &S) {text = strdup(S.text); return this;}
    ShortString operator= (const ShortString *S) {text = strdup(S->text); return this;}

    int operator== (const char *Text)            {return !(strcasecmp(text,Text));}
    int operator== (const ShortString &S)        {return !(strcasecmp(text,S.text));}
    int operator== (const ShortString *S)        {return !(strcasecmp(text,S->text));}

    int operator!= (const char *Text)            {return (strcasecmp(text,Text)!=0);}
    int operator!= (const ShortString &S)        {return (strcasecmp(text,S.text)!=0);}
    int operator!= (const ShortString *S)        {return (strcasecmp(text,S->text)!=0);}

    int MatchStart(const char *Text)     {return !(strncasecmp(text,Text, strlen(Text)));}
    int MatchStart(const ShortString &S) {return !(strncasecmp(text,S.text, strlen(S.text)));}
    int MatchStart(const ShortString *S) {return !(strncasecmp(text,S->text, strlen(S->text)));}

    char* operator() () {return text;}
    char  operator() (int i) {if (text!=NULL) return text[i]; else return -1;}

    operator char*() {return text;}     // auto type conversion

    float asFloat()  { if (text==NULL) return -1; 
                           return (strchr(text,'%'))?(atof(text)/100):atof(text); }
    //int asInt()    { if (text==NULL) return -1; return atoi(text); }
    int asInt()      { if (text==NULL) return -1; return (int) strtol(text, (char **)NULL, 0); } 

#ifdef PT
    pt asPoint()    
    {
        ShortString Pt = text; 
        float x=0;
        float y=0;
        float z=0; 
        char *ptr;
        ptr = strtok(Pt.text,"<>, ");

        if (ptr!=NULL) 
        {
            x = atof(ptr);
            ptr = strtok(NULL,"<>, ");
            if (ptr!=NULL)
            {
                y = atof(ptr);
                ptr = strtok(NULL,"<>, ");
                if (ptr!=NULL) 
                {
                    z = atof(ptr);
                }
            }
        }
        return pt(x,y,z);
    }  
#endif
    char* asString() { return text; } 
};


class Parser
{
public:
    ifstream InFile;

    virtual ~Parser() {}                  // to make the compiler happy and cut down on warnings...

    void ScreamAndDie(char *S, int i = 0)
    {
        cerr << S << " " << i << "\n";
        exit(0);
    }

    virtual void ParseFtn()
    {
    /*  e.g. : 
        while (InFile >> Word)
        {
            if (Word(0)=='#')
            {
                InFile->getline(Line, LINELEN);
            }
            else if (Word=="Bloop")
            {
                // handle Bloop
            }
            else 
            {
                // default case 
            } 
    */               
    }

    virtual void ParseCommand(char *S)
    {
        switch (S[0])
        {
             case 'h': 
             case 'H': HelpMsg(); break; 
        }
    }

    virtual void HelpMsg()
    {
        cout << "parse <infile> [-h]  i\n";
        cout << "This is intended as a base class - or as a null parser.\n";
        cout << "it reads in 'infile' and runs ParseFtn()\n"; 
    }

public:

    void Run(int argc, char *argv[])
    {
        if (argc==1) 
            ScreamAndDie("Doesn't handle piped input yet. Sorry. (-h for help)\n");

    	while (--argc>0)
	{
            if (argv[argc][0]=='-')
	        ParseCommand(&argv[argc][1]);
            else if (argc==1)
            {
                InFile.open(argv[1]);
                if (!InFile) 
                    ScreamAndDie(strcat("Unable to open input file ",argv[argc]));
            }
            else
                cerr<<"Unknown parameter " << argv[argc];
       }

       if (!InFile)
            ScreamAndDie("Doesn't handle piped input yet. sorry. (-h for help)\n");
       else
       {
            ParseFtn();
       }
    }
};

ostream& operator << (ostream &o, ShortString &S);
ostream& operator << (ostream &o, ShortString *S);
istream& operator >> (istream &i, ShortString &S);
istream& operator >> (istream &i, ShortString *S);

/*
ostream& operator << (ostream &o, ShortString &S)
{
    return o << S.text;
}

ostream& operator << (ostream &o, ShortString *S)
{
    return o << *S; 
}

istream& operator >> (istream &i, ShortString &S)
{
    return o >> S.text;
}

istream& operator >> (istream &i, ShortString *S)
{
    return o >> *S;
}

int main (int argc, char *argv[])
{
    new Filter(argc, argv);
    return (0);
}
*/

			
#endif
