#include <iostream.h>
#include <fstream.h>
#include <stdlib.h>
#include <math.h>

#include "sgi.h"
#include "point.h"
#include "polygon.h"


#ifdef DEBUG
#define ERRORLOG error_file
extern ofstream ERRORLOG;  
#else
#define ERRORLOG cout
#endif

polygon *polygon::Poly_no[MAXPOLYS];
int	 polygon::No_polys=0;

polygon::polygon(point& a, point& b, point& c)
// intialise a polygon with three points
{
    A=&a; B=&b; C=&c;
    normal = ~((b.P - a.P) * (c.P - a.P));
    if (No_polys<MAXPOLYS) 
	Poly_no[No_polys++] = this;
    else
	cerr << "\nWARNING!! Too many polygons";
}

polygon::polygon(int a, int b, int c)
// initialise a polygon from the global points list
{
    A = point::Point_no[a-1];
    B = point::Point_no[b-1];
    C = point::Point_no[c-1];

    normal = ((B->P - A->P) * (C->P - A->P));
    normal = ~normal;
    if (No_polys<MAXPOLYS) 
	Poly_no[No_polys++] = this;
    else
	cerr << "\nWARNING!! Too many polygons";
}

void polygon::show()
// This draws a polygon to a GL window: Requires such a window to already be opened.
{
    bgnpolygon();
	A->show();
	B->show();
	C->show();
    endpolygon();
}	

void polygon::Illuminate(pt I)
// This illuminates the polygon, given an Illumination source in direction I
// (Which is the apparent direction of illumination)
{
    pt N = ~(I*A->Vnorm);		//N is perpendicular to light and velocity
    double cos_alpha, alpha_dash;
    float speed = pt::abs(A->V);	//scaler speed of point
    I = ~I;				//normalise direction of light

    cos_alpha = I^A->Vnorm;
    alpha_dash = acos((cos_alpha + speed)/(1 + speed*cos_alpha));
    light = A->Vnorm*(cos(alpha_dash)) + ~(A->Vnorm*N)*(sin(alpha_dash)); 

    float Intensity = normal^light;		// A *simple* phong lighting Model
    if (Intensity>1) Intensity = 1;	 //Errr.... would you believe rounding error?
		 			 //... Also this allows for a saturating light 
					 //source (i.e. of strength > 1). - often happens 
					 //when the light source is being rapidly approached.
    if (Intensity<0) Intensity = 0;
    A->SetIntensity(Intensity);
    B->SetIntensity(Intensity);
    C->SetIntensity(Intensity);
};

void polygon::ResetNormal()
{
    normal = ((B->P - A->P) * (C->P - A->P));
    normal = ~normal;
}

void polygon::Show_all()
{
//ERRORLOG << "\nShow_all";
//    for (int i=0; i<No_polys; i++) Poly_no[i]->show();

    point *a,*b,*c;
    polygon *Old;
    long i;
    int Strip = FALSE;

    for (i=1; i<No_polys; i++)
    {
        Old = Poly_no[i-1];
        a = Old->A; b=Old->B; c=Old->C;

        if ((Poly_no[i]->A == b) && (Poly_no[i]->B == c))
        {
            if (!Strip)
            {
                Strip = TRUE;
                bgntmesh();
                a->show();
                b->show();
                c->show();
            }
            else
                c->show();
        }
        else if ((Poly_no[i]->A == a) && (Poly_no[i]->B==c))
        {
            if (!Strip)
            {
                Strip=TRUE;
                bgntmesh();
                a->show();
                b->show();
                swaptmesh();
                c->show();
            }
            else
            {
                swaptmesh();
                c->show();
            }
        }
        else
        {
            if (Strip)
            {

                c->show();      // finish the mesh off.
                endtmesh();
                Strip = FALSE;
            }
            else
                Old->show();    // show the whole polygon - no mesh
        }
    }

    // Draw the last polygon
    if (Strip)
    {
        if (Poly_no[No_polys-1]->A == a)
        {
                swaptmesh();
        }
        Poly_no[No_polys-1]->C->show();
        endtmesh();
    }
    else
        Poly_no[No_polys-1]->show();

}

void polygon::Init_all(pt light_source)
{
//ERRORLOG << "\nInit_all";
    for (int i=0; i<No_polys; i++) Poly_no[i]->Illuminate(light_source);
}
