#ifndef PT
#define PT

#define ANGLECONV	0.01745   
#define TRUE		1
#define FALSE		0
#define MAXPOINTS	100000
#define FPERROR         1e-5             // acceptable floating point error limit 
                                         // (for vector comparisons)
#define GROSSERROR      1e-3             // max f.p. error (for rough vector comparisons)
//#define DEBUG
/*
 *  pt - a normal, 3D point
 */

class pt {
public:
    double x, y, z;
    //pt()			    {x=0; y=0; z=0;}
    pt(double X=0, double Y=0, double Z=0)   {x=X; y=Y; z=Z;}

    pt operator= (const double A)    {x=A; y=A; z=A; return *this;}
    pt operator= (const pt p)       {x=p.x; y=p.y; z=p.z; return *this;}
    pt operator+ (const pt q)       {return pt(x+q.x, y+q.y, z+q.z);}
    pt operator/ (const double A)    {return pt(x/A, y/A, z/A);}
    pt operator* (const double A)    {return pt(x*A, y*A, z*A);}
    pt operator+= (const pt p)      {*this = *this + p; return(*this);}
    pt operator-= (const pt p)      {*this = *this - p; return(*this);}
    pt operator*= (const double A)   {*this = *this * A; return(*this);}
    pt operator/= (const double A)   {*this = *this / A; return(*this);}
    pt operator- (const pt p)	    {return pt(x-p.x, y-p.y, z-p.z);}
    pt operator- ()                 {return pt(-x, -y, -z);}

    //Cross Product
    pt operator* (const pt q)	    {return pt(y*q.z-z*q.y, z*q.x-x*q.z, x*q.y-y*q.x);}
    
    //Normalisation  
    pt operator~ ()		    {double S=sqrt(x*x+y*y+z*z);
                                     return((S==0)?pt(0,0,1):((S==1)?pt(x,y,z):pt(x/S,y/S,z/S)));}
    //dot product
    double operator^ (const pt p)    {return (x*p.x+y*p.y+z*p.z);}

    //Comparison with 0 (only!)
    int operator== (const double f)  {return( (f!=0) ? -1 : ((x==0)&&(y==0)&&(z==0)) );} 
    int operator!= (const double f)  {return( (f!=0) ? -1 : ((x!=0)||(y!=0)||(z!=0)) );} 

    pt parallel   (pt p)            {return (~p*(*this^(~p)));}

//    int operator== (pt p)           {return ((x==p.x)&&(y==p.y)&&(z==p.z));}
    int operator== (pt p)           {return ((x<p.x+FPERROR)&&(x>p.x-FPERROR)&&
                                             (y<p.y+FPERROR)&&(y>p.y-FPERROR)&&
                                             (z<p.z+FPERROR)&&(z>p.z-FPERROR));}

//    int operator!= (pt p)           {return ((x!=p.x)||(y!=p.y)||(z!=p.z));}
    int operator!= (pt p)           {return ((x>p.x+FPERROR)||(x<p.x-FPERROR)||
                                             (y>p.y+FPERROR)||(y<p.y-FPERROR)||
                                             (z>p.z+FPERROR)||(z<p.z-FPERROR));}
    
    static int aboutequal(const pt &a, const pt &b)     
                                    {return ((a.x<b.x+GROSSERROR)&&(a.x>b.x-GROSSERROR) &&
                                             (a.y<b.y+GROSSERROR)&&(a.y>b.y-GROSSERROR) &&
                                             (a.z<b.z+GROSSERROR)&&(a.z>b.z-GROSSERROR));} 

    pt perp       (pt p)            {return (*this-(parallel(p)));}

    double abs()                     {return(sqrt(x*x+y*y+z*z));} 


    //inverse
    pt inv()                        {return(pt(1/this->x,1/this->y,1/this->z));}
    static void invert(pt p)        {p.x = 1/p.x; p.y = 1/p.y; p.z = 1/p.z;}

    //Element-wise multiplication
    pt element_mult(pt p)         {return(pt(this->x*p.x,this->y*p.y,this->z*p.z));}    

    static double abs (const pt& a)  {return(sqrt(a.x*a.x+a.y*a.y+a.z*a.z));}

    static double fast_abs(const pt& a) { return fast_dist(a,a); }

    static double sqr (const pt &a)  {return(a.x*a.x+a.y*a.y+a.z*a.z);}

    static int   comp (pt &a, pt &b) {return((a.x>b.x-0.0001&&a.x<b.x+0.0001)&&
                                             (a.y>b.y-0.0001&&a.y<b.y+0.0001)&&
                                             (a.z>b.z-0.0001&&a.z<b.z+0.0001));}

    static int   comp (pt *a, pt *b) {return comp(*a, *b);}
 
    static double dist(pt a, pt b)    {return sqrt(((a.x-b.x)*(a.x-b.x))+
                                                  ((a.y-b.y)*(a.y-b.y))+
                                                  ((a.z-b.z)*(a.z-b.z)));}

    // This is a very simple minded but fast approximation to true distance
    // (maybe do a faster, all integer version using shifts?)
    // ALWAYS RETURNS LARGER THAN OR EQUAL TO TRUE DIST 
    static double fast_dist (const pt& a, const pt& b) 
    {
        double A = (a.x>b.x)?a.x - b.x:b.x-a.x;
        double B = (a.y>b.y)?a.y - b.y:b.y-a.y;
        double C = (a.z>b.z)?a.z - b.z:b.z-a.z;

/*      double temp; 
        if (B>A) 
        { 
            if (C>B) {temp = A; A=C; C=temp;}
            else if (C>A) {temp = A; A=B; B=C; C=temp;}
            else {temp = A; A=B; B=temp;}
        }
        else if (C>B) {temp = B; B=C; C=temp;}
            
        return (A+B/2+C/4); 
*/
//And now, for hack value:  (and maybe an infintesimal speed increase :-) )

        return ((A>B)?
                   ((B>C)? A+B/2+C/4 : A+C/2+B/4 )
               :  
                   ((A>C)?
                       B+A/2+C/4
                   : 
                       ((B>C)? B+C/2+A/4 : C+B/2+A/4)));
    }

// Hmmm... is this worth it, I wonder?

    static double fast_dist_low(const pt&a, const pt& b)
    {
        return fast_dist(a,b)*0.87;
    }

    void  Rotate (pt A, double a);
    // Takes a point and rotates it around an axis A (passing through (0,0,0))
    // by an angle a in radians.
/*    {

        pt r, N;

        if (A==0) return;

        r = (*this!=A);   //r = perpendicular component of this point
        N = (*this*(~A)); //normal to this point (vector) and the axis A

        //GO FASTER STRIPES
        static double aa; 
        static double Perp, Norm;

        if (aa!=a)
        {
            Perp = cos(a);
            Norm = sin(a); 
            aa   = a;
        }
        //END GO FASTER STRIPES

        r = r*Perp + N*Norm;
        *this = (*this==A) + r;
    }
*/


    //Used for angular momentum calculations - Omega gives axis and magnitude
    void Rotate(pt Omega)
    {
        Rotate(~Omega, abs(Omega));
    }

    static double abs (const pt& a, const pt& b)	
                                    {return (sqrt((a.x-b.x)*(a.x-b.x)+(a.y-b.y)*(a.y-b.y)+(a.z-b.z)*(a.z-b.z)));}
    static double sqr (const pt& a, const pt& b)	
                                    {return ((a.x-b.x)*(a.x-b.x)+(a.y-b.y)*(a.y-b.y)+(a.z-b.z)*(a.z-b.z));}

    static double sqr (pt *a, pt *b) {return pt::sqr((*a), (*b));} 

    friend ostream& operator << (ostream &o, pt &p);
    friend istream& operator >> (istream &i, pt &p);
};
    ostream& operator << (ostream &o, pt &p); 
    ostream& operator << (ostream &o, pt *p); 
    istream& operator >> (istream &i, pt &p);

/*Following code must be inserted somewhere useful in any program using 
  pt.h  - bizarre, but that's c++

ostream& operator << (ostream &o, pt &p)
{
    o<<"<"<<p.x<<","<<p.y<<","<<p.z<<">";
    return o;
}
*/

#endif
