#include "structure.h"

#define NOTDEFINED -999999

StructureData::StructureData(int noStructureObjects)
{
    ExclusionsDefined = FALSE;
    ExcludeLTX = NOTDEFINED;
    ExcludeGTX = NOTDEFINED;
    ExcludeLTY = NOTDEFINED;
    ExcludeGTY = NOTDEFINED;
    ExcludeLTZ = NOTDEFINED;
    ExcludeGTZ = NOTDEFINED;

    NoStructureObjects = noStructureObjects;
    Positions = new (PosData* [NoStructureObjects]);
    for (int i=0; i<NoStructureObjects; i++) Positions[i] = NULL;
    for (int i=0; i<MAXMODELS; i++) NoObjectsOfModelType[i] = 0;
cout << "Structure Constructed\n";
}

void StructureData::SetNoObjectsForModel(int Model, int No)
{
//cout << "Setting ["<<Model<<"] = " << No<<"\n";
    NoObjectsOfModelType[Model] = No;
}

int StructureData::NoType(int Model)
{
   return NoObjectsOfModelType[Model]; 
}

void StructureData::SetData(int Index, PosData* P)
{
    if (Index>=NoStructureObjects)
    {
        cout << " Error in Structure File\n"; 
        cout << " (attempt to add PosData object outside Model No Limits: "<<Index<<")\n";
    } 
    else
    {
        if (P==NULL)
            cout << "Error : setting index ["<<Index<<"] to a NULL PosData value!\n";
        Positions[Index] = P;
    }
} 

PosData* StructureData::GetData(int Model, int Offset)
{
    int index = 0;
    for (int i=0; i<Model; i++) index += NoType(i);

    index += Offset;
    if (index < NoStructureObjects)
    {
        if (Positions[index]==NULL)
        {
            cout << " index " << index << " not occupied by anything ("<<Model<<","<<Offset<<")\n";
        }
        return Positions[index];
    }
    else
    {
        cout << "Error: attempt to access out of bounds PosData "<<index<<"\n";
        return NULL;
    }
}

void StructureData::SetExclusionPlane(pt Plane, int ExcludeNegative)
{
    if (Plane == 0) return;

    ExclusionsDefined = TRUE;

    if (Plane.x != 0)
    {
        if (ExcludeNegative) 
            ExcludeLTX = Plane.x;
        else
            ExcludeGTX = Plane.x; 
    }
    else if (Plane.y != 0)
    {
        if (ExcludeNegative)
            ExcludeLTY = Plane.y;
        else
            ExcludeGTY = Plane.y;
    }
    else if (Plane.z != 0)
    {
        if (ExcludeNegative)
            ExcludeLTZ = Plane.z;
        else
            ExcludeGTZ = Plane.z;
    }
    else 
    {
        cout << " Wierd error parsing exclusion plane: "<<Plane<<"\n";
    }
}

void StructureData::print()
{
    cout << "Structure Print Out\n";
    for (int i=0; i<NoStructureObjects; i++)
    {
        cout << " Pos Data # "<<i<<" ";

        if (Positions[i]==NULL)
            cout << "NULL\n";
        else
            Positions[i]->print();
    }
}
