/*
 *   A data class for holding information about pre-existing structures
 *   that the simulation will start up using.  It is read from a
 *   structure data file, that must use the same names for objects as
 *   defined in the .pddf info file.
 *
 *   The structure data file is simply a list of positions and links,
 *   used to initialise objects at the start of the simulation, and
 *   also an optional exclusion zone designed as a plane or series
 *   of planes.
 */

#ifndef STRUCTURE
#define STRUCTURE

#include "tubular.h"
#include "matrix.h" 

class Microtubule;

/*
 *    PosData contains all the information needed to initialise an object
 *    to a certain position, direction and state.
 *
 *    Purely a data class. No methods.
 */

class PosData
{
    public:

    int ModelNo;
    pt center;                  // the co-ordinate center of the object
    int State;                  // Current State of object.
    Microtubule *Owner;         // the microtubule this dimer is part of (if any).
    int bound;                  // whether the object is bound or free
    CBMatrix Basis;             // The basis vectors, expressed as a matrix

    PosData() {};
    PosData(int Mod_no, pt center_pos, CBMatrix basis_matrix, int state) 
        {ModelNo = Mod_no; center=center_pos; Basis=basis_matrix; State = state;}

    void print() {cout << ModelNo<<" "<<center<<"\n";} 
};

class StructureData
{
    public:

    int NoStructureObjects;

    PosData** Positions;

    int NoObjectsOfModelType[MAXMODELS];

    int ExclusionsDefined;

    float ExcludeLTX,ExcludeGTX,ExcludeLTY,ExcludeGTY,ExcludeLTZ,ExcludeGTZ;

    StructureData(int noStructureObjects);

    void SetNoObjectsForModel(int Model, int No);

    int NoType(int Model);   // return no. of objects of model type M

    void SetData(int Index, PosData* P);

    PosData* GetData(int Model, int Offset); 

    void SetExclusionPlane(pt Plane, int ExcludeNegative = TRUE);

    void print();
};
#endif  STRUCTURE
