/* 
 *   See bottom of file for Program Doco
 *
 *    "When I wrote this code, both I and God knew how it
 *     worked.  Now only God knows."
 */


#include "tubular.h"
#include "pt.h"
#include "surfette.h"
#include "dimer.h"
#include "gridcell.h"
#include "testbench.h"


/*
void OutputSummary();
void ParseCommandLine();
void PrintHelp(char *topic);
void PrintParserHelp();
void PrintVersion();
*/

// #define DATALOG data_files 
ofstream DATALOG1;  
ofstream DATALOG2;
ofstream DATALOG3;
ofstream DATALOG4;
ofstream DATALOG5;

// #define ERRORLOG error_file 
ofstream ERRORLOG;  

#define DEBUGCYCLE FALSE 

Program TheProgram;

/*
ostream& operator << (ostream &o, pt &p)
{
    o<<"<"<<p.x<<","<<p.y<<","<<p.z<<">";
    return o;
}
*/

Program::Program()
{
    collision_test     = FALSE;
    collision_test_moving = FALSE;

    number_of_dimers   = DEF_NUMB_DIMERS;
    field_size         = DEF_FIELD_SIZE;
    half_field_size    = DEF_FIELD_SIZE/2;
    int_half_field_size= DEF_FIELD_SIZE/2;
    half_field_size_sq = (DEF_FIELD_SIZE*DEF_FIELD_SIZE/4);
    freeze_mts         = FALSE;
    width              = 512;
    height             = 512;
    cycle_length       = CALCS_PER_FRAME;
    povrayfreq         = 0;
    tubrepfreq         = -1;
    rgbfreq            = 0;
    molarity           = 0;
    silent             = FALSE;
    verbose            = FALSE;
    iterations         = -1;
    NoCyclesCompleted  = 0;
    NoGraphics         = FALSE;
    NoRotation         = FALSE;
    NoMultLinkBreakage = FALSE;
    ParseOnly          = FALSE;
    NoMerge            = FALSE;
 
    PDDFString         = NULL;
    StructureString    = NULL;
    TestSequence       = FALSE;
    TimeScale          = 1;

    for (int i=0; i<=10000; i++)  
       SqrtArray[i] = sqrt(i); 
} 

void Program::PrintVersion()
{
    if (!silent) cout << "tubular - version 1.10     [May 1997]\n";
    if (!silent) cout << "   author - Chris Betts\n";
}

void Program::OutputSummary()
{
    if (molarity>50) cout << "\nWARNING: Physically Improbable Molarity.\n";

//    DATALOG << "# molarity (microMoles): " << molarity << " ("
//        <<number_of_dimers<< " dimers )\n";

    if (TestSequence)
        DATALOG1 << "# max no. of iterations per test cycle  " << iterations<<"\n";
//    DATALOG << "# binding/unbinding strengths (T B S t b s) [numbt mts] [totalmass] [avweight] [turbidity]\n"; 


    cout <<"\nparameters: \n";
    cout <<"  field size: "<<field_size<<" nanometers cubed\n";
    cout <<"  molarity (microM) : "<<molarity<<" ("<<number_of_dimers<<" dimers)";
    cout << "\n   updates every " << cycle_length << " frames \n";

    if (silent) return;
    
}


void Program::PrintHelp(char *topic)
{
    if (silent) return;

    if (strcasecmp(topic,"parser")==0) {PrintParserHelp();return;}
    if (strcasecmp(topic,"graphics")==0) {PrintGraphicsHelp();return;}

    cout << "\nnanosim - a nanoscale molecular self-assembly simulator\n";
    cout << "\n options:";
    cout << "\n  -B            -> no breakage of multiple links"; 
    cout << "\n  -c[Integer]   -> number of computation cycles between visual updates";
    cout << "\n  -C[+/-]       -> Test no of collision for non-binding object motile('+'), non('-')";
    cout << "\n  -d[Integer]   -> save data logs every N iterations";
    cout << "\n  -D[File Name] -> Data: enter name for PDDF Data File (default is 'info.pddf')"; 
    cout << "\n  -f[Integer]   -> size of simulation field in microns";
    cout << "\n  -F            -> Freeze microtubule movement";
    cout << "\n  -G            -> Suppresses graphical output";
    cout << "\n  -h/-H         -> this help message ";
    cout << "\n  -Hgraphics    -> help on Graphics Display/Interaction";
    cout << "\n  -m[Integer]   -> the number of monomer objects to use";
    cout << "\n  -M[Integer]   -> set molarity in micromoles ";
    cout << "\n  -NM           -> debug: no merging of aggregates"; 
    cout << "\n  -p[Integer]   -> save povray files every N iterations"; 
    cout << "\n  -P            -> run the parser only, don't run simulation"; 
    cout << "\n  -r[Integer]   -> save sgi `rgb' files every N iterations"; 
    cout << "\n  -R            -> Microtubules don't rotate";
    cout << "\n  -s            -> Silent: suppresses all text output (except errors)";
    cout << "\n  -S            -> Data: name of Structure Data file (default is 'info.struct')"; 
    cout << "\n  -T[Integer]   -> only run sim for N iterations"; 
    cout << "\n  -V            -> Verbose: produces lots of text output"; 
    cout << "\n  -v            -> Print Version number only"; 
    cout << "\n  -x[Integer]   -> width of Graphics window (default 512)";
    cout << "\n  -y[Integer]   -> height of Graphics window (default 512)";

    cout << "\n";
}

/*
 *   Prints help about the parser and parser file.  Things like the units
 *   used etc. might be useful here...
 */

void Program::PrintParserHelp()
{
    cout << "See the example file example.pddf for details of the parser\n";
}

/*
 *  Prints help about useing the (primative) interactivity of the graphics
 *  view window, and saving graphics files.
 */

void Program::PrintGraphicsHelp()
{
    cout << "\n The graphics Window supports some simple functions;";
    cout << "\n\n the mouse will move the viewpoint around"; 
    cout << "\n The following keys have special functions";
    cout << "\n [R] - refresh the view";
    cout << "\n [Q] - quit the entire program";
    cout << "\n [B] - toggle showing only Bound objects off/on";
    cout << "\n [P] - pause the simulation";
    cout << "\n [S] - save the current image as an RGB file";
    cout << "\n [V] - save the current image as a POVray file";
    cout << "\n\n";
}

/*
 * Iterate throught the command line parameters.  
 * Some of them overwrite others - it tries to ensure consistency.
 * (for readability, and curious aesthetics, tab has become 2 spaces here...) 
 */

int Program::ParseCommandLine(int argc, char *argv[])
{ 
  int DimersSet = FALSE;
  int MolaritySet = FALSE;

  while (--argc)
  {
    switch(argv[argc][1])
    {
     case 'B': NoMultLinkBreakage = TRUE; break; 

     case 'c': cycle_length     = atoi(argv[argc]+2); break; 

     case 'C': collision_test   = TRUE;                      
                                  if (strlen(argv[argc]+1)>1)
                                  {
                                      if (argv[argc][2] == '+')
                                          collision_test_moving = TRUE;
                                      else if (argv[argc][2] == '-')
                                          collision_test_moving = FALSE;
                                      else
                                          cerr <<"unknown collision test option "<<argv[argc]+2<<"\n";
                                  } 
                                  break;

     case 'm': number_of_dimers = atoi(argv[argc]+2); 
                                  DimersSet = TRUE;
                                  break;

     case 'd': tubrepfreq       = atoi(argv[argc]+2); break;

     case 'D': PDDFString       = argv[argc]+2; break;

     case 'f': field_size       = atoi(argv[argc]+2); 
                                  break;
 
     case 'F': freeze_mts       = TRUE; NoRotation = TRUE;  break;

     case 'h': case 'H':        PrintHelp(argv[argc]+2);         return EXIT; 

     case 'G': NoGraphics       = TRUE; break;     

     case 'M': molarity         = atof(argv[argc]+2); 
                                  MolaritySet = TRUE;
                                  break;

     case 'N': NoMerge          = TRUE; break;  

     case 'p': povrayfreq       = atoi(argv[argc]+2); break;
           
     case 'P': ParseOnly         = TRUE; break;
           
     case 'r': rgbfreq          = atoi(argv[argc]+2); break;

     case 'R': NoRotation       = TRUE; break;
           
     case 'S': StructureString  = argv[argc]+2; 
                                  if (strlen(StructureString)==0) StructureString="info.struct"; 
                                  break;

     case 's': silent           = TRUE; break;

     case 'T': iterations       = atoi(argv[argc]+2); break;

     case 'V': verbose          = TRUE; break;
 
     case 'v':                  return EXIT;
           
     case 'x': width            = atoi(argv[argc]+2); break;

     case 'y': height           = atoi(argv[argc]+2); break;
 
     case 'z':                  TestSequence = TRUE; break;

     default: cerr << " Unknown Parameter '" << argv[argc] <<"'\n"; 
    }
  }

   //Convert between molarity and number_of_dimers, depending on which
   //the user has given us.  (number_of_dimers is the actual variable
   //used by the program, molarity is a user interface luxury.)


  if (field_size != DEF_FIELD_SIZE)
  {
      if (molarity == 0 && (MolaritySet==FALSE) && (DimersSet==FALSE)) 
          molarity = DEF_MOLARITY; //may want 0 moles for structure tests
      field_size = CheckFieldSize(field_size);
      half_field_size    = field_size/2;
      int_half_field_size = (int) half_field_size;
      half_field_size_sq = field_size*field_size/4;
  }

  if (molarity>0 || MolaritySet) 
      number_of_dimers=(int)(pow((float)field_size,3)*4.8E-6*molarity);
  else
      molarity=((float)number_of_dimers/(pow((float)field_size,3)*4.8E-6));
 
  return SanityCheck();
}

/*
 *   Checks that field size is sane, and is a power of two.
 *   *cough*: there's almost certainly a neater way to do this... :-)
 */

int Program::CheckFieldSize(int field_size)
{
    int test = field_size;

    if (test < 1)
    {
        cout << " ERROR: field size must be a positive integer (not "<<test<<").\n"; 
        cout << " Setting to 256\n";
        field_size = 256;
    }
    else
    {
        while (test != 1)
        {
            if ((test%2) != 0)
            {
                cout << " ERROR: field size must be a power of two. ";
                cout << " (was " << field_size << ") \n";
                int max = 1; 
                while (test>max) {max *=2;} 
                field_size = max;
                test = 1;
                cout << "   Resetting field size to : " << field_size << "\n";
            }
            else
                test/=2;
        }
    }
    return field_size;
}

int Program::SanityCheck()
{
    // Check that things are sane

    if (number_of_dimers < 0)
    {
         cout << " Error - bad number of objects " << number_of_dimers << " - setting";
         cout << " to 1 object\n";
         number_of_dimers = 1;
    }
    else if (number_of_dimers == 0)
    {
         cout << " Warning - 0 free objects specified.  Simulation will be fairly boring.\n";
    } 
 
    return OK;
}

int main(int argc, char *argv[])
{
    return (TheProgram.Run(argc, argv));
}

void Program::OpenStream(ofstream* o, char* fileName, int modifier, int timestamp)
{
    o->open(fileName, modifier);
    if (!(&o)) { cerr << "\nFATAL ERROR: Unable to open log file '" << fileName << "\n";  exit(-1); } 

    if (timestamp!=FALSE)
    {
        if (!TheProgram.silent) 
            cout << "# log file " << fileName << " opened\n"; 
        time_t bloop = time(NULL);
        *o << "# file " << fileName << " opened at " << ctime(&bloop) <<"\n";
    }

}

void Program::CloseStream(ofstream* o, int timestamp)
{
    if (timestamp != FALSE)
    {
cout << "close timestamp = " << timestamp << " " << FALSE << "\n";
        time_t bloop = time(NULL);
        (*o) << "# file closed " << ctime(&bloop) << "\n";
    }
    o->close();
}

// so it's ugly.  Sue me.

void Program::OpenLogs(int modifier, int timestamp)
{
if (timestamp!=FALSE) cout << "OPENING LOGS!!\n";
    OpenStream(&DATALOG1, "tubdata1.log", modifier, timestamp);
    OpenStream(&DATALOG2, "tubdata2.log", modifier, timestamp);
    OpenStream(&DATALOG3, "tubdata3.log", modifier, timestamp);
    OpenStream(&DATALOG4, "tubdata4.log", modifier, timestamp);
    OpenStream(&DATALOG5, "tubdata5.log", modifier, timestamp);
    OpenStream(&ERRORLOG, "tuberror.log", modifier, timestamp);
}

void Program::CloseLogs(int timestamp)
{
    CloseStream(&DATALOG1, timestamp); 
    CloseStream(&DATALOG2, timestamp); 
    CloseStream(&DATALOG3, timestamp);
    CloseStream(&DATALOG4, timestamp); 
    CloseStream(&DATALOG5, timestamp); 
    CloseStream(&ERRORLOG, timestamp);
}

/*
 * Umm... flush the buffers and close + reopen files so
 * that the logs are still around even if the program
 * crashes or is externally terminated...
 * (There's probably a better way to do this...)
 */

void Program::CloseAndReopenLogs()
{
    CloseLogs(FALSE);
    OpenLogs(ios::app, FALSE); 
}

int Program::Run(int argc, char *argv[])
{
    PrintVersion();

    // Init fast normal random number generator
    long Time = time(NULL);
    initnorm(Time);

    if (ParseCommandLine(argc, argv)==EXIT)
        return 0;

    OpenLogs(ios::out, TRUE);

    // CONSIDER - putting field_size sanity check in here somewhere...
    
    OutputSummary();

    ERRORLOG << "Dimer initialisation\n";

    GridCell::Init(field_size);

    cout << "number_of_dimers = " << number_of_dimers << "\n";

    if (tubrepfreq==-1) tubrepfreq = cycle_length;

    if (ParseOnly)   // Test Code
    {
        TubParser* Parser = new TubParser();
        Parser->parse((PDDFString==NULL)?"info.pddf":PDDFString);
        cout<<"\nParser Test only (-P option).  Quitting.\n";
        return 0;
    }

    TubParser* Parser = (Dimer::Init(number_of_dimers, field_size, PDDFString, StructureString));
    if (Parser == NULL)
    {
        cerr << "Error: unable to initialise with " << number_of_dimers;
        cerr << " dimers and a fields size of " << field_size << "\n"; 
        cerr << " (molarity "<<molarity<<"micromoles is far too high)\n";
        return -1;
    }

    TheProgram.Temperature = Parser->Temperature;
    TheProgram.Viscosity   = Parser->Viscosity;
    TheProgram.TimeScale   = Parser->TimeScale;

    if (!TheProgram.NoGraphics)
        Surf::InitWindow(width,height, field_size, Parser);

    if (TestSequence==TRUE || collision_test == TRUE)
        return RunTestSequence();
    else
    {
        CycleSim();
        LogDataRun();
    }

    CloseLogs(TRUE);

    cout << "Normal Program Termination\n";
    return 0;
}

int Program::RunTestSequence()
{
//    TestBench::RunTest();


    int fin = FALSE;
    if (iterations < 0) iterations = 100;
    int TestIterations = iterations;

    if (collision_test == TRUE)
    {
cout << "collision test moving = " << collision_test_moving << "\n";
        if (collision_test_moving == FALSE)  // test against non-moving dimer...
            Dimer::Dimers[0].rms_dp = 0;
        CycleSim();
        cout <<"         --------------           \n";
        cout <<"         -Test Results-           \n";
        cout <<"         --------------           \n";

       if (collision_test_moving == FALSE)
            cout << " Test object *not* moving \n";
       else
            cout << " Test object *is* moving \n";

        cout <<"No of Collisions in "<<TestIterations<<" steps (" << (TestIterations*TimeScale);
        cout << " seconds) = \n "<<Dimer::NoCollisions<<"\n"; 
        DATALOG1 <<"No of Collisions in "<<TestIterations<<" steps (" << (TestIterations*TimeScale);
        DATALOG1 << " seconds) = \n "<<Dimer::NoCollisions<<"\n"; 
        return OK;
    }
    else
    {
        while (!fin)
        {   
            //Every cycle reset things back to the starting state
            Dimer::ClearAll();
            Dimer::InitRandomBindStrengths(); 
            NoCyclesCompleted = 0;
            iterations = TestIterations; 

            fin = CycleSim();  

            DATALOG1 << setw(5) << NoCyclesCompleted << " "; 
            LogDataRun();
        }
    }
    time_t bloop = time(NULL);
    DATALOG1 << "# DATALOG Experiment Ended: " << ctime(&bloop) << "\n"; 
    cout << "Experimental Sequence Ended " << ctime(&bloop)<<"\n";


    return OK;
}
void Program::LogDataRun()
{
/*
    DATALOG << setprecision(4) << setiosflags(ios::showpoint | ios::fixed);
    DATALOG << setw(5) << Dimer::bindtop <<" ";  
    DATALOG << setw(5) << Dimer::bindbottom <<" ";  
    DATALOG << setw(5) << Dimer::bindside <<" ";  
    DATALOG << setw(5) << Dimer::unbindtop <<" ";  
    DATALOG << setw(5) << Dimer::unbindbottom <<" ";  
    DATALOG << setw(5) << Dimer::unbindside <<" ";  
    DATALOG << setw(5) << Microtubule::count <<" ";
    DATALOG << setw(5) << Microtubule::totalmass <<" ";
    DATALOG << setw(5) << Microtubule::avweight <<" ";
    DATALOG << setw(5) << Microtubule::turbidity << "\n";
*/
}


int Program::CycleSim()
{

    if (!TheProgram.silent) ERRORLOG << "Start Cycling simulation...\n";

    int Finished = FALSE;
    int UserFinished = FALSE;
    long cycles = cycle_length;
    long frameno = 0;
    int Time = time(NULL);
    if (!TheProgram.silent) ERRORLOG << "Time is: " << Time << "\n";
    srandom(Time);
    
    while (Finished == FALSE)
    {
        Microtubule::CheckAll("Start Cycle Check: ");

        if ((rgbfreq>0) && (cycles%rgbfreq==0))
             Surf::Write_next_rgb_file();

        if ((povrayfreq>0) && (frameno%povrayfreq==0))
             Surf::Write_next_pov_file();

        if (cycles++>=cycle_length-1)
        { 
            UserFinished = Surf::view();
if (UserFinished) cout << " caught user finished from view()\n";
            Dimer::CheckAll("Occasional Check");            
//            Microtubule::Report();
//            Dimer::Report();
            cycles = 0;
        }

/*
 *  Note on grouping logic:  
 *
 *     Things to do with aggregates - run by microtubule class
 * 
 *     Things to do with individual objects - run by dimer class
 *
 *     Things to do with object interaction/spatial data structure
 *         - run by gridcell class
 */


        Dimer::Check_binding();

        Dimer::Random_events();

        GridCell::CheckAllBounds();     // make sure dimers haven't jiggled too far

        Microtubule::MoveAll();         // jiggle the microtubules 

        Dimer::Agitate();               // move the dimers around randomly

        GridCell::CheckAllBounds();     // make sure dimers haven't jiggled to far

        GridCell::LoadCells();          // set up the grid of dimers

        GridCell::GlobalInteraction();  // bind close tubulin dimers together

        GridCell::CheckSpacing();       // check there are no collisions

        Dimer::Integrity_check();       // Make sure everything is o.k. 

        Microtubule::CheckAll("Post Checks: ");

        if ((tubrepfreq>0) && (cycles%tubrepfreq==0))
        {
             Microtubule::Report();
             Dimer::Report();
             CloseAndReopenLogs();
        }

        frameno++;
     
        if (UserFinished == FALSE)  // check again...
            UserFinished = Surf::CheckUserInput();
        Finished = UserFinished;
        NoCyclesCompleted++;
        if (iterations>=0)
        {
            if (iterations--==0) Finished = TRUE;
        }

    } 
    long secs = (time(NULL)-Time);
    cout << "Time (secs) " << secs;
    cout << " Speed Index (KDC/s)) = " << float(NoCyclesCompleted*number_of_dimers)/(float(1000)*secs)<<"\n";

    cout << "\nNormal Cycle Termination\n";

    Microtubule::FinalReport();

    return UserFinished;
}

/*
 *   Following are some utility functions to do with the simulation field 
 *   boundaries.
 */

/*
 *  BoundsOffset() and OffsetCoOrds()  
 *
 *  OffsetCoOrd() takes a float, and calculates the offset 
 *  needed to "wrap around" the co-ordinate back into the simulation
 *  region (i.e. -half_field_size to +half_field_size).
 *
 *  BoundsOffset() takes a pt, and calculates an offset that, if
 *  added to the pt, would move it back into the field.
 *
 *        - returns a zero pt if no offset is required
 *
 *   USAGE:
 *        this is where it gets tricky.  Often this is used
 *        to bring back a 'stray' dimer, that wonders off field.
 *        (it returns an offset which, added to the dimer, brings
 *        the dimer to an appropriate pos. on the other side.)
 *
 *        It is also sometimes used (a little trickily) in reverse,
 *        to move a dimer *out* of the boundary fields in a
 *        particular direction, esp. in the microtubule moment
 *        of inertia calcs, and gridcell interaction calcs.
 */


int Program::OffsetCoOrd(float x)
{
    int I_Pos;
    int X = (int)x;

    if (x>0)
    {
        I_Pos = ((X+int_half_field_size)%field_size) - int_half_field_size; 
    }
    else
    {
        I_Pos = ((X-int_half_field_size)%field_size) + int_half_field_size; 
    }

    return (I_Pos - X);
}

pt Program::BoundsOffset(pt P)
{
   pt offset(0);

   if (P.x>=half_field_size || P.x<-half_field_size) offset.x = OffsetCoOrd(P.x);
   if (P.y>=half_field_size || P.y<-half_field_size) offset.y = OffsetCoOrd(P.y);
   if (P.z>=half_field_size || P.z<-half_field_size) offset.z = OffsetCoOrd(P.z);
   
   return offset;
}

/*
 *  returns a point that has been moved back to within the program bounds region.
 */

pt Program::MoveIntoBounds(pt P)
{
   pt New = P + BoundsOffset(P);
   return New;
}

// checks that two points are equal, when wrapping is taken into account

int Program::EqualWithWrap(pt A, pt B)
{
    pt A2 = MoveIntoBounds(A);
    pt B2 = MoveIntoBounds(B);    
    return (pt::aboutequal(A2,B2)); 
}


/*
 *   Fastish (I hope :-) ) sqrt routine accurate to at 
 *   worst 0.5%, on an "average" 0.025% (?)  
 *   uses a scaled table lookup.  Fully Accurate for all
 *   numbers which are 3 digit decimals (i.e. 412, or .0733)
 *
 *   TODO: check speed, remove recursion for loop?
 */

float Program::FastishSqrt(float X)
{
#ifdef USEFASTMATHS

    if (X>=100&&X<9999)
    {
        return fastsqrt(X);
    }
    else if (X<100&X>=1)
    {
        return FastishSqrt(X*100)/10;
    }
    else if (X<1&&X>0.01)
    {
        return FastishSqrt(X*10000)/100;
    }
    else
        return sqrt(X);
#else
    return sqrt(X);
#endif
}

/*
pt Program::RandomNormalPt()
{
    pt NormPt(FastGauss, FastGauss, FastGauss);
    assert(isnan(NormPt)==FALSE);
    return NormPt;
}
*/

/*
 *  Checks if a pt is within the simulation field
 */

int Program::OutOfBounds(pt testpt)
{
    if (testpt.x>half_field_size||testpt.x<-half_field_size||testpt.y>half_field_size||
        testpt.y<-half_field_size||testpt.z>half_field_size||testpt.z<-half_field_size)
    {
        return TRUE;
    }
    else
    {
        return FALSE;
    }
}

/*

   Tubular is a small program that does a simple simulation of the way
tubulin monomers combine to form microtubules.

   The basic plot is to have a whole lot of double-spheres,
dumb-bell like, floating in simulated bliss.  A number of interaction 
and movement factors are defined for these double spheres ('dimers'), which 
come in three types: alpha tubulin, beta tubulin (phosphorylated) and beta
tubulin (unphosphorylated).  Forward/Backwards and sideways linkage  
strengths are also defined.  After initial conditions are set, the  
simulation is left alone, and the dimers coagulate... hopefully into
something  microtubule like.

    A number of classes are defined to let us do this:

    'pt'  : defined in pt.h   a simple 3d point class, defining a lot
                              of common vector operations.

    'dimer' :  dimer.cc/.h    defines the dimer, and its attractive
                              forces.  A number of class methods
                              define the interactions of large groups
                              of dimers.

    'surfette': surfette.cc/.h    a cut down SGI viewing class, vaguely
                                  optimised for fast display.
   
    'polygon': polygon.cc/.h  A polygon class, used in the surfette
                              class.

    'gridcell': gridcell.cc/.h   A 'virtual space management' class, used
                                 to optimise the handling of interactions
                                 between large numbers of dimers (avoiding
                                 N**2 interactions per step).  It works by
                                 defining a 3d grid of dimer 'buckets', with
                                 interactions only being checked within a 
                                 bucket and between adjoining buckets.


*/ 
